Package: scanMiRApp
Type: Package
Title: scanMiR shiny application
Version: 1.19.0
Date: 2026-02-25
Authors@R: c(
    person("Pierre-Luc", "Germain", email="pierre-luc.germain@hest.ethz.ch", 
      role=c("cre","aut"), comment=c(ORCID="0000-0003-3418-4218")),
    person("Michael", "Soutschek", email="michael.soutschek@hest.ethz.ch", role="aut"),
    person("Fridolin", "Gross", email="fridolin.gross@hest.ethz.ch", role="ctb"))
Imports: AnnotationDbi, AnnotationFilter, AnnotationHub, BiocParallel,
        Biostrings, data.table, digest, DT, ensembldb, fst,
        GenomeInfoDb, GenomicFeatures, GenomicRanges, ggplot2,
        htmlwidgets, IRanges, Matrix, methods, plotly, rintrojs,
        rtracklayer, S4Vectors, scanMiRData, shiny, shinycssloaders,
        shinydashboard, shinyjqui, stats, utils, txdbmaker, waiter
Suggests: knitr, rmarkdown, BiocStyle, testthat (>= 3.0.0), shinytest,
        BSgenome.Hsapiens.UCSC.hg38, BSgenome.Mmusculus.UCSC.mm10,
        BSgenome.Mmusculus.UCSC.mm39, BSgenome.Rnorvegicus.UCSC.rn6
Description: 
    A shiny interface to the scanMiR package. The application enables the 
    scanning of transcripts and custom sequences for miRNA binding sites, the 
    visualization of KdModels and binding results, as well as browsing predicted
    repression data. In addition contains the IndexedFst class for fast indexed
    reading of large GenomicRanges or data.frames, and some utilities for 
    facilitating scans and identifying enriched miRNA-target pairs.
Depends: R (>= 4.0), scanMiR
License: GPL-3
VignetteBuilder: knitr
RoxygenNote: 7.2.3
biocViews: miRNA, SequenceMatching, GUI, ShinyApps
Config/testthat/edition: 3
git_url: https://git.bioconductor.org/packages/scanMiRApp
git_branch: devel
git_last_commit: 7b6d92e
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-29 00:52:26 UTC; biocbuild
Author: Pierre-Luc Germain [cre, aut] (ORCID:
    <https://orcid.org/0000-0003-3418-4218>),
  Michael Soutschek [aut],
  Fridolin Gross [ctb]
Maintainer: Pierre-Luc Germain <pierre-luc.germain@hest.ethz.ch>

# Generated by roxygen2: do not edit by hand

export(ScanMiRAnno)
export(enrichedMirTxPairs)
export(getTranscriptSequence)
export(loadIndexedFst)
export(plotSitesOnUTR)
export(runFullScan)
export(saveIndexedFst)
export(scanMiRApp)
export(scanMiRserver)
export(scanMiRui)
exportClasses(IndexedFst)
exportClasses(ScanMiRAnno)
exportMethods("$")
exportMethods("[")
exportMethods("[[")
exportMethods(as.data.frame)
exportMethods(colnames)
exportMethods(head)
exportMethods(length)
exportMethods(lengths)
exportMethods(names)
exportMethods(ncol)
exportMethods(nrow)
exportMethods(show)
exportMethods(summary)
import(Biostrings)
import(GenomicRanges)
import(IRanges)
import(Matrix)
import(methods)
import(scanMiR)
import(shiny)
import(shinydashboard)
importFrom(AnnotationDbi,select)
importFrom(AnnotationFilter,AnnotationFilter)
importFrom(AnnotationFilter,AnnotationFilterList)
importFrom(AnnotationFilter,SeqNameFilter)
importFrom(AnnotationHub,AnnotationHub)
importFrom(AnnotationHub,query)
importFrom(BiocParallel,MulticoreParam)
importFrom(BiocParallel,SerialParam)
importFrom(BiocParallel,bpnworkers)
importFrom(Biostrings,DNAString)
importFrom(Biostrings,DNAStringSet)
importFrom(Biostrings,RNAString)
importFrom(DT,DTOutput)
importFrom(DT,datatable)
importFrom(DT,renderDT)
importFrom(GenomeInfoDb,seqlevels)
importFrom(GenomicFeatures,"seqlevels<-")
importFrom(GenomicFeatures,cds)
importFrom(GenomicFeatures,cdsBy)
importFrom(GenomicFeatures,exons)
importFrom(GenomicFeatures,exonsBy)
importFrom(GenomicFeatures,extractTranscriptSeqs)
importFrom(GenomicFeatures,threeUTRsByTranscript)
importFrom(GenomicRanges,GRanges)
importFrom(GenomicRanges,makeGRangesFromDataFrame)
importFrom(GenomicRanges,strand)
importFrom(IRanges,IRanges)
importFrom(S4Vectors,"mcols<-")
importFrom(S4Vectors,"metadata<-")
importFrom(S4Vectors,mcols)
importFrom(S4Vectors,metadata)
importFrom(data.table,rbindlist)
importFrom(digest,digest)
importFrom(ensembldb,"seqlevelsStyle<-")
importFrom(ensembldb,exonsBy)
importFrom(ensembldb,genes)
importFrom(ensembldb,getGenomeTwoBitFile)
importFrom(ensembldb,metadata)
importFrom(ensembldb,seqlevelsStyle)
importFrom(ensembldb,threeUTRsByTranscript)
importFrom(ensembldb,transcripts)
importFrom(fst,fst.metadata)
importFrom(fst,read.fst)
importFrom(fst,threads_fst)
importFrom(fst,write.fst)
importFrom(ggplot2,aes)
importFrom(ggplot2,aes_string)
importFrom(ggplot2,element_line)
importFrom(ggplot2,expand_limits)
importFrom(ggplot2,geom_hline)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_rect)
importFrom(ggplot2,geom_text)
importFrom(ggplot2,geom_vline)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,labs)
importFrom(ggplot2,scale_x_continuous)
importFrom(ggplot2,scale_y_continuous)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_light)
importFrom(ggplot2,theme_minimal)
importFrom(ggplot2,xlab)
importFrom(ggplot2,xlim)
importFrom(htmlwidgets,JS)
importFrom(methods,show)
importFrom(plotly,event_data)
importFrom(plotly,event_register)
importFrom(plotly,ggplotly)
importFrom(plotly,plotlyOutput)
importFrom(plotly,renderPlotly)
importFrom(rintrojs,hintjs)
importFrom(rintrojs,introBox)
importFrom(rintrojs,introjs)
importFrom(rintrojs,introjsUI)
importFrom(rintrojs,readCallback)
importFrom(rtracklayer,TwoBitFile)
importFrom(rtracklayer,export.2bit)
importFrom(scanMiRData,getKdModels)
importFrom(shinycssloaders,withSpinner)
importFrom(shinyjqui,jqui_resizable)
importFrom(stats,pbinom)
importFrom(stats,setNames)
importFrom(txdbmaker,makeTxDbFromGRanges)
importFrom(utils,capture.output)
importFrom(utils,data)
importFrom(utils,object.size)
importFrom(utils,packageVersion)
importFrom(utils,write.csv)
importFrom(waiter,spin_1)
importFrom(waiter,use_waiter)
importFrom(waiter,waiter_hide)
importFrom(waiter,waiter_show)
importFrom(waiter,waiter_show_on_load)
importMethodsFrom(methods,show)

<img align="right" style="margin-left: 25px; margin-bottom: 30px;" src="https://raw.githubusercontent.com/ETHZ-INS/scanMiR/master/inst/docs/sticker.svg"/>

# scanMiRApp

`scanMiRApp` is a companion package to <a href="https://github.com/ETHZ-INS/scanMiR">scanMiR</a>.
It contains:

* A shiny interface to the <a href="https://github.com/ETHZ-INS/scanMiR">scanMiR</a> package,
enabling the scanning of transcripts (or custom sequences) for miRNA binding sites, the 
visualization of KdModels and binding results, as well as browsing predicted repression data.
* The `ScanMiRAnno` class, which encapsulates all the annotation information necessary for the
app/wrappers in this package for a given species.
* The <a href="vignettes/IndexedFST.Rmd">`IndexedFst`</a> class for fast indexed reading of 
large GenomicRanges or data.frames; this is used to enable fast random access to pre-compiled 
scans and aggregations without having to load them in memory.
* A number of convenience wrappers to `scanMiR` functions, for tasks like scanning or
visualizing sites on the UTR of a given transcript, running transcriptome-wide scans, or 
finding enriched miRNA-target pairs.

Click [here](https://ethz-ins.org/scanMiR/) for a live version of the app.
For more information, see the package's [vignettes](vignettes/) and the [preprint](https://doi.org/10.1101/2021.06.16.448293).

## Installation

First ensure that <a href="https://github.com/ETHZ-INS/scanMiR">scanMiR</a> and 
<a href="https://github.com/ETHZ-INS/scanMiRData">scanMiRData</a> are installed, then install 
with:

```{r}
BiocManager::install("ETHZ-INS/scanMiRApp")
```

#' @exportClass ScanMiRAnno
#' @rdname ScanMiRAnno
setClass(
  "ScanMiRAnno",
  contains="list",
  validity=function(object){
    stopifnot(!is.null(object$genome) &&
              (is(object$genome, "BSgenome") || is(object$genome, "TwoBitFile")))
    stopifnot(!is.null(object$ensdb) &&
                (is(object$ensdb, "EnsDb") || is(object$ensdb, "TxDb")))
    if(!is.null(object$models))
      stopifnot(is(object$models, "KdModelList") ||
                  is(object$models, "character"))
    if(!is.null(object$scan))
      stopifnot(is(object$scan, "GRanges") ||
                  is(object$scan, "IndexedFst"))
    if(!is.null(object$aggregated))
      stopifnot(is(object$aggregated, "data.frame") ||
                  is(object$aggregated, "IndexedFst"))
  }
)

#' ScanMiRAnno
#'
#' @param species The species/build acronym for automatic construction; if
#' omitted, `genome` and `ensdb` should be given. Current possible values are:
#' GRCh38, GRCm38, GRCm39, Rnor_6.
#' @param genome A \link[BSgenome]{BSgenome-class}, or a
#' \code{\link[rtracklayer]{TwoBitFile}}
#' @param ensdb An \link[ensembldb]{EnsDb-class} (or a
#' \link[GenomicFeatures]{TxDb-class}) object
#' @param models An optional KdModelList
#' @param scan An optional full scan (IndexedFst or GRanges)
#' @param aggregated An optional per-transcript aggregation (IndexedFst or
#' data.frame)
#' @param version optional ensembl version
#' @param addDBs A named list of additional tx-miRNA databases, each of which
#' should be a data.frame with the columns 'transcript', 'miRNA', and 'score'.
#' @param ... Arguments passed to `AnnotationHub`
#'
#' @return A `ScanMiRAnno` object
#' @export
#' @importFrom AnnotationHub AnnotationHub query
#' @importFrom ensembldb getGenomeTwoBitFile
#' @importFrom scanMiRData getKdModels
#' @examples
#' anno <- ScanMiRAnno(species="fake")
#' anno
ScanMiRAnno <- function(species=NULL, genome=NULL, ensdb=NULL, models=NULL,
                        scan=NULL, aggregated=NULL, version=NULL,
                        addDBs=list(), ...){
  blink <- NULL
  if(!is.null(species)){
    species <- match.arg(species, c("GRCh38","GRCm38","GRCm39","Rnor_6","fake"))
    if(species=="fake") return(.fakeAnno())
    if(is.null(models))
      models <- switch(species,
                       GRCh38=scanMiRData::getKdModels("hsa"),
                       GRCm38=scanMiRData::getKdModels("mmu"),
                       GRCm39=scanMiRData::getKdModels("mmu"),
                       "Rnor_6"=scanMiRData::getKdModels("rno"),
                       #"Rnor_7"=scanMiRData::getKdModels("rno"),
                       stop("Species not among the pre-defined one, please",
                            "provide `models` manually.")
                      )
    ah <- AnnotationHub(...)
    if(is.null(ensdb))
      ensdb <- ah[[rev(query(ah, c("EnsDb", species, version))$ah_id)[1]]]
    genome <- switch(species,
      GRCh38=BSgenome.Hsapiens.UCSC.hg38:::BSgenome.Hsapiens.UCSC.hg38,
      GRCm38=BSgenome.Mmusculus.UCSC.mm10:::BSgenome.Mmusculus.UCSC.mm10,
      GRCm39=BSgenome.Mmusculus.UCSC.mm39:::BSgenome.Mmusculus.UCSC.mm39,
      "Rnor_6"=BSgenome.Rnorvegicus.UCSC.rn6:::BSgenome.Rnorvegicus.UCSC.rn6,
      #"Rnor_7"=BSgenome.Rnorvegicus.UCSC.rn7:::BSgenome.Rnorvegicus.UCSC.rn7,
      getGenomeTwoBitFile(ensdb)
    )
    blink <- switch( species,
      GRCh38="https://www.ensembl.org/Homo_sapiens/Gene/Summary?db=core;g=",
      GRCm38="https://nov2020.archive.ensembl.org/Mus_musculus/Gene/Summary?db=core;g=",
      GRCm39="https://www.ensembl.org/Mus_musculus/Gene/Summary?db=core;g=",
      "Rnor_6"="https://www.ensembl.org/Rattus_norvegicus/Gene/Summary?db=core;g=",
      NULL )
  }
  new("ScanMiRAnno", list(
    genome=genome, ensdb=ensdb, models=models, scan=scan, aggregated=aggregated,
    addDBs=addDBs, ensembl_gene_baselink=blink
  ))
}


#' Methods for the \code{\link{ScanMiRAnno}} class
#' @name ScanMiRAnno-methods
#' @rdname ScanMiRAnno-methods
#' @aliases ScanMiRAnno-methods
#' @seealso \code{\link{ScanMiRAnno}}
#' @param object An object of class \code{\link{ScanMiRAnno}}
#' @return Depends on the method.
NULL

#' @rdname ScanMiRAnno-methods
#' @aliases ScanMiRAnno-methods
#' @export
setMethod("summary", "ScanMiRAnno", function(object){
  show(object$genome)
  cat("\n")
  show(object$ensdb)
  cat("\nModels:\n")
  if(is.null(object$models)){
    cat("None\n")
  }else{
    summary(object$models)
  }
  cat("\nScan:\n")
  if(is.null(object$scan)){
    cat("None\n")
  }else{
    summary(object$scan)
  }
  cat("\nAggregated:\n")
  if(is.null(object$aggregated)){
    cat("None\n")
  }else{
    summary(object$aggregated)
  }
  if(!is.null(object$addDBs) && length(object$addDBs)>0){
    cat("\nAdditional DBs: ", paste(names(object$addDBs),collapse=", "), "\n")
  }
})

#' @rdname ScanMiRAnno-methods
#' @aliases ScanMiRAnno-methods
#' @export
#' @importFrom methods show
setMethod("show", "ScanMiRAnno", function(object){
  if(is(object$genome, "BSgenome")){
    gm <- paste0(metadata(object$genome)$organism, " (",
                 metadata(object$genome)$genome, ")")
  }else{
    gm <- toString(object$genome)
  }
  em <- setNames(metadata(object$ensdb)$value, metadata(object$ensdb)$name)
  if(is(object$ensdb,"EnsDb")){
    em <- paste0(em[["Organism"]], " (", em[["genome_build"]],") v",
                 em[["ensembl_version"]])
  }else{
    em <- paste0(em[["Organism"]], " (", em[["Genome"]],")")
  }
  cat("Genome:", gm, "\nAnnotation:", em)
  if(!is.null(object$models)) cat("\nModels:", class(object$models),
                                  "of length", length(object$models))
  if(!is.null(object$scan)) cat("\n + Scan")
  if(!is.null(object$aggregated)) cat("\n + Aggregated")
  cat("\n")
})


#' @importFrom Biostrings DNAStringSet
#' @importFrom rtracklayer TwoBitFile export.2bit
#' @importFrom txdbmaker makeTxDbFromGRanges
#' @importFrom utils data
.fakeAnno <- function(){
  data("SampleKdModel", package="scanMiR", envir = environment())
  data("SampleTranscript", package="scanMiR", envir = environment())
  f <- tempfile()
  export.2bit(DNAStringSet(SampleTranscript), f)
  ge <- TwoBitFile(f)
  gr <- GRanges(seqlevels(ge), IRanges(c(1,1,1,1),c(898,898,898,210)),
                type=c("gene","transcript","exon","CDS"),
                exon_id=c(NA,NA,"e1",NA), exon_number=c(NA,NA,1,NA),
                strand="+", source="fake", gene_biotype="protein_coding")
  gr$entrezid <- gr$gene_id <- gr$gene_name <- "gene1"
  gr$transcript_id <- gr$tx_id <- "ENSTFAKE0000056456"
  md <- data.frame(name=c("Organism","Genome"), value=c("Fake falsus","fake1"))
  db <- txdbmaker::makeTxDbFromGRanges(gr, metadata=md)
  ScanMiRAnno(genome=ge, ensdb=db, models=c(SampleKdModel))
}

suppressPackageStartupMessages({
  library(shiny)
  library(shinytest)
  library(scanMiRApp)
})

anno <- ScanMiRAnno("fake")

test_that("shiny app (server) works", {
  skip_on_cran()
  testServer(scanMiRserver(list(fake=anno)), {
    session$setInputs(gene="gene1", transcript="ENSTFAKE0000056456",
                      mirlist="fake", seqFeature="3' UTR only",
                      subjet_type="transcript", annotation="fake", shadow=15L,
                      keepmatchseq=FALSE, minDist=7L, maxLogKd=-1L,
                      scanNonCanonical=TRUE, circular=FALSE)
    session$setInputs(mirnas="hsa-miR-155-5p", mirnas_all=FALSE)
    expect_equal(selgene(),"gene1")
    expect_equal(seltx(),"ENSTFAKE0000056456")
    expect_equal(names(selmods()), "hsa-miR-155-5p")
    cs <- checksum()
    cached.hits[[checksum()]] <- do.scan()
    current.cs(cs)
    session$flushReact()
    h <- hits()$hits
    expect_equal(GenomicRanges::start(h), c(281L,482L))
    expect_equal(as.character(h$type), c("8mer","7mer-m8"))
    p <- output$manhattan
    selectedMatch(h[1])
    session$flushReact()
    expect(grepl("||||||||||||     ||||||||", output$alignment, fixed=TRUE),
           "viewTargetAlignment doesn't seem to work as expected", info)

  })
})

#' runFullScan
#'
#' Runs a full miRNA scan on all protein-coding transcripts (or UTRs) of an
#' annotation.
#'
#' @param annotation A \code{\link{ScanMiRAnno}} object
#' @param mods An optional `KdModelList` (defaults to the one in `annotation`)
#' @param annoFilter An optional `AnnotationFilter` or `AnnotationFilterList` to
#' filter the set of transcripts to be extracted
#' @param extract Which parts of the transcripts to extract. For `UTRonly`
#' (default) only the 3' UTR regions are extracted, `withORF` additionally
#' extracts the coding regions, and `exons` extracts all exons
#' @param shadow The size of the ribosomal shadow at the UTR starts
#' @param cores The number of threads to use. Alternatively accepts a
#' \code{\link[BiocParallel]{BiocParallelParam-class}}, as for instance produced
#' by \code{\link[BiocParallel]{MulticoreParam}}.
#' @param maxLogKd The maximum log_kd of sites to report
#' @param save.path Optional, the path to which to save the results
#' @param onlyCanonical passed to \code{\link[scanMiR]{findSeedMatches}}
#' @param ... Arguments passed to \code{\link[scanMiR]{findSeedMatches}}
#'
#' @return A `GRanges` object
#'
#' @export
#' @import Biostrings scanMiR
#' @importFrom BiocParallel SerialParam MulticoreParam bpnworkers
#' @importFrom GenomicFeatures extractTranscriptSeqs threeUTRsByTranscript
#' exonsBy cdsBy
#' @importFrom GenomicRanges strand
#' @importFrom S4Vectors metadata metadata<-
#' @importFrom stats setNames
#' @importFrom AnnotationFilter AnnotationFilterList SeqNameFilter AnnotationFilter
#' @examples
#' anno <- ScanMiRAnno("fake")
#' m <- runFullScan( annotation=anno )
#' m
runFullScan <- function(annotation, mods=NULL, annoFilter = NULL,
                        extract=c("UTRonly", "withORF", "exons"),
                        onlyCanonical=TRUE, shadow=15, cores=1,
                        maxLogKd=c(-1,-1.5), save.path=NULL, ...){

  stopifnot(is(annotation, "ScanMiRAnno"))
  if(is.null(mods)) mods <- annotation$models
  if(is(mods,"KdModel")) mods <- KdModelList(mods)
  stopifnot(is(mods,"KdModelList"))

  if(is.numeric(cores) && length(cores)==1){
    cores <- as.integer(cores)
    if(cores>1){
      BP <- MulticoreParam(cores, progressbar=TRUE)
    }else{
      BP <- SerialParam()
    }
  }else if(is(cores,"BiocParallelParam")){
    BP <- cores
  }else{
    warning("`cores` argument of unknown type -- will be ignored.")
    BP <- SerialParam()
  }

  message("Loading annotation")
  genome <- annotation$genome
  ensdb <- annotation$ensdb
  if(is(ensdb,"EnsDb")) seqlevelsStyle(genome) <- "Ensembl"

  # restrict to canonical chromosomes
  canonical_chroms <- seqlevels(genome)[!grepl('_', seqlevels(genome))]
  filt <- SeqNameFilter(canonical_chroms)

  if(!is.null(annoFilter)){
    if(!is(annoFilter, "AnnotationFilterList") &&
       !is(annoFilter, "AnnotationFilter"))
      stop("filter must be either `AnnotationFilter` or `AnnotationFilterList`")
    filt <- AnnotationFilterList(filt, annoFilter)
    if(!is(ensdb,"EnsDb"))
      warning("`annoFilter` is ignored when the annotation is not in ",
              "`EnsDb` format", immediate.=TRUE)
  }else{
    if(!is(ensdb, "EnsDb")) filt <- NULL
  }

  message("Extracting transcripts")
  seqs <- getTranscriptSequence(tx=NULL, annotation, annoFilter=filt,
                                extract=extract)

  message("Scanning with ", bpnworkers(BP), " thread(s)")
  m <- findSeedMatches(seqs, mods, shadow=shadow, maxLogKd=maxLogKd,
                       onlyCanonical=onlyCanonical, BP=BP, verbose=FALSE, ...)

  md <- metadata(annotation$ensdb)
  md <- setNames(md$value,md$name)
  if(!("genome_build" %in% names(md)))
    md[["genome_build"]] <- md[["Genome"]]
  if(!("ensembl_version" %in% names(md))) md[["ensembl_version"]] <- NA

  if(is(m, "GRanges")) {
    metadata(m)$genome_build <- md[["genome_build"]]
    metadata(m)$ensembl_version <- md[["ensembl_version"]]
  } else {
    attr(m, "ensembl_version") <- md[["ensembl_version"]]
    attr(m, "genome_build") <- md[["genome_build"]]
  }
  if(is.null(save.path)) return(m)
  saveRDS(m, file=save.path)
  rm(m)
  gc()
  message("Saved in: ", save.path)
}

#' scanMiRserver
#'
#' Server function for the scanMiR shiny app. Most users are expected to use
#' \code{\link{scanMiRApp}} instead.
#'
#' @param annotations A named list of \code{\link{ScanMiRAnno}} object.
#' @param modlists A named list of `KdModelList` objects. If omitted, will
#' fetch it from the annotation objects.
#' @param gc.time The number of microseconds after which a garbage collect
#' should be performed.
#' @param maxCacheSize Maximum cache size in bytes.
#' @param BP BPPARAM for multithreading
#'
#' @return A shiny server function
#' @importFrom digest digest
#' @importFrom BiocParallel SerialParam
#' @importFrom plotly renderPlotly ggplotly event_data event_register
#' @importFrom ensembldb genes transcripts exonsBy threeUTRsByTranscript
#' @importFrom DT datatable renderDT DTOutput
#' @importFrom htmlwidgets JS
#' @importFrom digest digest
#' @importFrom shinycssloaders withSpinner
#' @importFrom ggplot2 ggplot aes_string geom_hline geom_point expand_limits
#' xlab geom_vline geom_rect theme_minimal theme element_line scale_x_continuous
#' scale_y_continuous
#' @importFrom AnnotationDbi select
#' @importFrom Biostrings DNAStringSet RNAString DNAString
#' @importFrom waiter waiter_hide waiter_show
#' @importFrom rintrojs hintjs introjs readCallback
#' @importFrom utils capture.output object.size write.csv packageVersion
#' @importFrom S4Vectors mcols mcols<-
#' @import shiny shinydashboard scanMiR GenomicRanges IRanges
#' @export
#' @examples
#' # we'd normally fetch a real annotation:
#' # anno <- ScanMiRAnno("Rnor_6")
#' # here we'll use a fake one:
#' anno <- ScanMiRAnno("fake")
#' srv <- scanMiRserver(list(fake=anno))
scanMiRserver <- function( annotations=list(), modlists=NULL, gc.time=Inf,
                           maxCacheSize=10*10^6, BP=SerialParam() ){
  stopifnot(length(annotations)>0)
  stopifnot(all(vapply(annotations, class2="ScanMiRAnno",
                       FUN.VALUE=logical(1), FUN=is)))
  if(is.null(modlists))
    modlists <- lapply(annotations, FUN=function(x) x$models)
  stopifnot(all(vapply(modlists, class2="KdModelList",
                       FUN.VALUE=logical(1), FUN=is)))
  stopifnot(all(names(modlists) %in% names(annotations)))

  dtwrapper <- function(d, pageLength=25, rownames=TRUE, ...){
    datatable( d, filter="top", class="compact",
               extensions=c("Buttons","ColReorder"),
               options=list(
                 pageLength=pageLength, dom = "fltBip", colReorder=TRUE,
                 buttons=c('copy', 'csv', 'excel', 'csvHtml5', 'colvis')
               ), rownames=rownames, ... )
  }

  checkModIdentity <- function(m1,m2){
    identical(lapply(m1,FUN=function(x) x$mer8),
              lapply(m2,FUN=function(x) x$mer8))
  }

  getTxs <- function(db, gene=NULL){
    if(is.null(gene)) return(NULL)
    if(is(db,"EnsDb")){
      if(!is.null(gene)) filt <- ~gene_id==gene
      tx <- transcripts(db, columns=c("tx_id","tx_biotype"),
                        filter=~gene_id==gene, return.type="data.frame")
    }else{
      tx <- suppressMessages(try(select(db, keys=gene, keytype="GENEID",
                                    columns=c("TXNAME","TXTYPE")), silent=TRUE))
      if(is(tx,"try-error")) return(NULL)
      colnames(tx) <- c("gene","tx_id","tx_biotype")
    }
    if(nrow(tx)==0) return(NULL)
    setNames(tx$tx_id, paste0(tx$tx_id, " (",tx$tx_biotype, ")"))
  }

  getGeneFromTx <- function(db, tx){
    if(is(db,"EnsDb")){
      tx <- transcripts(db, columns=c("tx_id","gene_id"),
                        filter=~tx_id==tx, return.type="data.frame")
      return(as.character(tx$gene_id[1]))
    }
    tx <- suppressMessages(try(select(db, keys=tx, keytype="TXNAME",
                                      columns=c("TXID","GENEID")), silent=TRUE))
    if(is(tx,"try-error")) return(NULL)
    as.character(tx$GENEID[1])
  }

  function(input, output, session){

    observe({
      if(gc.time>0 & !is.infinite(gc.time)){
        invalidateLater(gc.time, session)
        gc(verbose=FALSE)
      }
    })

    #############################
    ## intro

    startIntro <- function(session){
      introjs(session, options=list(steps=.getAppIntro(), "nextLabel"="Next",
                                    "prevLabel"="Previous"),
              events=list(onbeforechange=readCallback("switchTabs")))
    }
    
    observeEvent(input$helpBtn, startIntro(session))
    observeEvent(input$helpLink, startIntro(session))

    ##############################
    ## initialize inputs

    output$menuCollection <- renderUI({
      menuItem(tags$span("miRNA Collection:", 
                         HTML("<br/>&nbsp;&nbsp;&nbsp;&nbsp;"), input$mirlist),
               tabName="tab_collection")
    })
    output$menuMirnas <- renderUI({
      if( (nsel <- length(selmods()))== 0){
        lab <- "miRNAs (none selected)"
      }else{
        lab <- paste0("miRNAs (", nsel, ")")
      }
      menuSubItem(lab, "tab_mirnas")
    })
    
    updateSelectizeInput(session, "mirlist", choices=names(modlists))
    updateSelectizeInput(session, "annotation", choices=names(annotations))

    observe({
      # when the choice of collection changes, update the annotation to
      # use the same genome
      if(!is.null(input$mirlist) && input$mirlist!="")
        updateSelectizeInput(session, "annotation", selected=input$mirlist)
    })


    ##############################
    ## select collection

    allmods <- reactive({ # all models from collection
      if(is.null(input$mirlist)) return(NULL)
      modlists[[input$mirlist]]
    })

    # prints a summary of the model collection
    output$collection_summary <- renderPrint({
      if(is.null(input$mirlist) || is.null(annotations[[input$mirlist]]))
        return(NULL)
      summary(allmods())
      ad <- annotations[[input$mirlist]]$addDBs
      if(!is.null(ad) && length(ad)>0)
        cat("\nAdditional DB(s):", paste(names(ad), collapse=", "))
    })
    output$selected_collection <- renderValueBox({
      if(is.null(input$mirlist) || is.null(annotations[[input$mirlist]]))
        return(valueBox("N/A", subtitle = "Nothing loaded", color="light-blue"))
      valueBox(input$mirlist, color = "light-blue",
        tags$div(lapply(capture.output(print(annotations[[input$mirlist]])),
               FUN=function(x) tags$p(x)))
      )
    })


    observe({ ## when the selected collection changes,
              ## update the miRNA selection inputs
      updateSelectizeInput(session, "mirnas", choices=names(allmods()),
                           server=TRUE)
      updateSelectizeInput(session, "mirna", choices=names(allmods()),
                           server=TRUE)
    })

    ##############################
    ## scan specific sequence

    ## transcript selection

    sel_ensdb <- reactive({ # the ensembldb for the selected genome
      if(is.null(input$annotation) || input$annotation=="" ||
         !(input$annotation %in% names(annotations))) return(NULL)
      annotations[[input$annotation]]$ensdb
    })

    allgenes <- reactive({ # all genes in the selected genome
      if(is.null(sel_ensdb())) return(NULL)
      if(is(sel_ensdb(), "EnsDb")){
        sl <- gsub("^chr","",seqlevels(annotations[[input$annotation]]$genome))
        g <- genes(sel_ensdb(), columns="gene_name", return.type="data.frame",
                   filter=SeqNameFilter(sl))
        gs <- setNames(g[,2], paste(g[,1], g[,2]))
      }else{
        g <- genes(sel_ensdb())
        names(gs) <- gs <- g$gene_id
        if(!is.null(g$gene_name)){
          names(gs) <- paste(g$gene_name, gs)
        }
      }
      gs
    })

    selgene <- reactive({ # selected gene id
      if(is.null(input$gene) || input$gene=="") return(NULL)
      input$gene
    })

    output$gene_link <- renderUI({
      if(is.null(selgene()) || selgene()=="") return(NULL)
      base <- annotations[[input$annotation]]$ensembl_gene_baselink
      if(is.null(base)) return(NULL)
      tags$a(href=paste0(base, selgene()), icon("external-link"),
             "view on ensembl", target="_blank")
    })

    alltxs <- reactive({ # all tx from selected gene
      if(is.null(selgene()) || selgene()=="") return(NULL)
      getTxs(sel_ensdb(), selgene())
    })

    seltx <- reactive({ # the selected transcript
      if(is.null(input$transcript) || input$transcript=="" ||
         is.na(input$transcript))
        return(NULL)
      changeFlag()
      input$transcript
    })

    # when the ensembldb is updated, update the gene input
    observe(updateSelectizeInput(session, "gene", choices=allgenes(),
                                 server=TRUE))
    # when the gene selection is updated, update the transcript input
    observe({
      prev_seltx <- input$transcript
      if(!(prev_seltx %in% alltxs())) prev_seltx <- NULL
      txs <- alltxs()
      if(is.null(txs)) txs <- c()
      updateSelectizeInput(session, "transcript", choices=alltxs(), selected=prev_seltx)
    })

    # takes a genome package name as input, and returns the genome
    getGenome <- function(x){
      if(is.character(x)){
        x <- library(x, character.only=TRUE)
        x <- get(x)
      }else if(is(x,"ScanMiRAnno")){
        x <- x$genome
      }
      seqlevels(x) <- gsub("^chr","",seqlevels(x))
      x
    }

    seqs <- reactive({ # returns the selected sequence(s)
      if((is.null(selgene()) || selgene()=="") &&
         (is.null(seltx()) || seltx()=="")) return(DNAStringSet())
      gid <- selgene()
      if(is.null(txid <- seltx())) txid <- getTxs(sel_ensdb(), gid)
      getTranscriptSequence( txid, annotations[[input$annotation]],
                             extract=switch(input$seqFeature,
         "CDS+UTR"="withORF", "whole transcript"="exons", "UTRonly"))
    })

    output$tx_overview <- renderTable({ # overview of the selected transcript
      if(is.null(seqs()) || length(seqs())==0)
        return(data.frame(sequence="Empty sequence!"))
      w <- width(seqs())
      ss <- as.character(subseq(seqs(), 1, end=ifelse(w<40,w,40)))
      ss[w>40] <- paste0(ss[w>40],"...")
      data.frame(transcript=names(seqs()), length=w, sequence=ss)
    })

    ## end transcript selection

    ## custom sequence

    customTarget <- reactive({
      if(is.null(input$customseq) || input$customseq=="") return(NULL)
      isRNA <- grepl("U", input$customseq, fixed=TRUE)
      seq <- gsub("[^ACGTUN]","", toupper(input$customseq))
      if(input$circular) seq <- paste0(seq,substr(seq,1,min(nchar(seq),11)))
      if(isRNA) seq <- RNAString(seq)
      DNAString(seq)
    })

    output$custom_info <- renderPrint({ # overview of the custom sequence
      if(is.null(input$customseq)) return("")
      out <- capture.output(customTarget())
      if(input$circular)
        out <- c("Circularized sequence: the first 11nt are pasted to ",
                 "the end of the sequence.\n", out)
      cat(out)
    })

    target <- reactive({ # target subject sequence
      if(input$subjet_type=="custom"){
        return(as.character(DNAStringSet(customTarget())))
      }
      if(is.null(seqs()) || length(seqs())>1) return(NULL)
      changeFlag()
      as.character(seqs())
    })

    observeEvent(input$rndseq, { # generate random sequence
      updateTextAreaInput(session, "customseq",
                          value=as.character(getRandomSeq()))
    })

    ## Select miRNAs for scanning

    observeEvent(input$mirnas_confident, {
      if(is.null(allmods())) return(NULL)
      cons <- conservation(allmods())
      if(all(is.na(cons))) return(NULL)
      updateSelectizeInput(session, "mirnas",
                           selected=names(cons)[as.numeric(cons)>1])
    })
    observeEvent(input$mirnas_mammals, {
      if(is.null(allmods())) return(NULL)
      cons <- conservation(allmods())
      if(all(is.na(cons))) return(NULL)
      updateSelectizeInput(session, "mirnas",
                           selected=names(cons)[as.numeric(cons)>2])
    })
    observeEvent(input$mirnas_vert, {
      if(is.null(allmods())) return(NULL)
      cons <- conservation(allmods())
      if(all(is.na(cons))) return(NULL)
      updateSelectizeInput(session, "mirnas",
                           selected=names(cons)[as.numeric(cons)>3])
    })
    observeEvent(input$mirnas_clear, {
      updateSelectizeInput(session, "mirnas", selected="")
    })
    
    observe({
      if(!is.null(input$mirnas) && length(input$mirnas)>0)
        updateCheckboxInput(session, "mirnas_all", value=FALSE)
    })
    observe({
      if(input$mirnas_all)
        updateSelectizeInput(session, "mirnas", selected="")
    })

    selmods <- reactive({ # models selected for scanning
      if(is.null(allmods())) return(NULL)
      if(input$mirnas_all) return(allmods())
      if(is.null(input$mirnas)) return(NULL)
      allmods()[input$mirnas]
    })

    ## Begin scan and results caching

    output$scanBtn <- renderUI({
      if(is.null(target()) || !isTRUE(nchar(target())>0) ||
         is.null(selmods()) || length(selmods())==0)
        return(actionButton("noscan", "Cannot launch scan - check input",
                            icon("exclamation-triangle"), disabled=TRUE))
      actionButton("scan", "Scan!", icon = icon("search"))
   })

    # actual and past scanning results are stored in this object
    cached.hits <- reactiveValues()

    changeFlag <- reactiveVal(0)

    cached.checksums <- reactive({
      ch <- reactiveValuesToList(cached.hits)
      ch <- ch[!vapply(ch, FUN.VALUE=logical(1), FUN=is.null)]
      ch <- lapply(ch, FUN=function(x){
        x[c("target","size","time","last","nsel","sel")]
      })
    })
    current.cs <- reactiveVal()

    hits <- reactive({ # the results currently loaded are stored in this object
      if(is.null(current.cs())) return(NULL)
      if(current.cs() %in% names(cached.checksums()))
        return(cached.hits[[current.cs()]])
      NULL
    })

    checksum <- reactive({ # generate a unique hash for the given input
      changeFlag()
      paste( digest::digest(selmods()),
             digest::digest(list(target=target(), shadow=input$shadow,
                                 keepMatchSeq=input$keepMatchSeq,
                                 minDist=input$minDist, maxLogKd=input$maxLogKd,
                                 scanNonCanonical=input$scanNonCanonical))
      )
    })

    cache.size <- reactive({
      ch <- cached.checksums()
      if(is.null(ch) || length(ch)==0) return(0)
      sum(vapply(ch, FUN.VALUE=numeric(1), FUN=function(x) as.numeric(x$size)))
    })

    cleanCache <- function(){
      # remove last-used results when over the cache size limit
      cs <- isolate(cached.checksums())
      if(length(cs)<3 || as.numeric(cache.size())<maxCacheSize) return(NULL)
      cs <- cs[order(unlist(lapply(cs,FUN=function(x) x$last)),decreasing=TRUE)]
      sizes <- vapply(ch, FUN.VALUE=numeric(1), FUN=function(x)
                                                            as.numeric(x$size))
      while(length(cs)>2 & sum(sizes)>maxCacheSize){
        cached.hits[[rev(names(cs))[1]]] <- NULL
        cs <- cs[-length(cs)]
        sizes <- sizes[-length(sizes)]
      }
    }

    checkPreComputedScan <- function(txid, utr_only=FALSE){
      if(is.null(preScan <- annotations[[input$annotation]]$scan)) return(NULL)
      if(is.null(origMods <- annotations[[input$annotation]]$models) ||
         !all(names(selmods()) %in% names(origMods)) ||
         !checkModIdentity(selmods(), origMods[names(selmods())])){
        warning("The models are not the same as those present in the ",
                "annotation; ignoring pre-compiled scans.")
        return(NULL)
      }
      if(is(preScan, "GRanges")){
        if(!(txid %in% seqlevels(preScan))){
          warning("Transcript not found in the pre-compiled scan!")
          return(NULL)
        }
        h <- preScan[seqnames(preScan)==txid &
                       preScan$miRNA %in% names(selmods())]
      }else{
        if(!(txid %in% names(preScan))){
          warning("Transcript not found in the pre-compiled scan!")
          return(NULL)
        }
        h <- preScan[[txid]]
        h <- h[h$miRNA %in% names(selmods())]
      }
      if(utr_only && !is.null(h$ORF)){
        h <- h[!h$ORF]
        h$ORF <- NULL
        if(!is.null(metadata(h)$tx_info)){
          s <- metadata(h)$tx_info[as.character(seqnames(h)), "ORF.length"]
          h <- IRanges::shift(h, -1L * as.integer(s))
        }
      }
      if(length(h)==0) return(h)
      if(!input$scanNonCanonical)
        h <- h[grep("canonical|bulged",h$type,invert=TRUE)]
      h <- h[h$log_kd < input$maxLogKd]
      h[order(h$log_kd)]
    }

    observeEvent(input$scan, { # actual scanning
      if(is.null(selmods()) || is.null(target()) || nchar(target())==0)
        return(NULL)
      cs <- checksum()
      cached.hits[[cs]] <- do.scan()
      current.cs(cs)
    })

    do.scan <- reactive({
      if(is.null(selmods()) || is.null(target()) || length(target())==0 ||
        nchar(target())==0){
        waiter_hide()
        return(NULL)
      }
      tmp <- changeFlag()
      cs <- checksum()
      # first check if we already have these results cached:
      if(cs %in% names(cached.checksums())) return(cached.hits[[cs]])
      waiter_show(color = "#333E4850")
      # then check if the results are pre-computed
      if(input$subjet_type!="custom" && input$seqFeature!="whole transcript" &&
         !is.null(h <- checkPreComputedScan(seltx(), input$seqFeature=="3' UTR only"))){
        res <- list(hits=h)
      }else{
        res <- list()
        msg <- paste0("Scanning sequence for ",length(selmods())," miRNAs")
        message(msg)
        detail <- NULL
        if(length(selmods())>4) detail <- "This might take a while..."
        if(input$circular)
          detail <- "'Ribosomal Shadow' is ignored when scanning circRNAs"
        scantarget <- target()
        scanmods <- selmods()
        keepmatchseq <- input$keepmatchseq
        shadow <- ifelse(input$circular,0,input$shadow)
        onlyCanonical <- !input$scanNonCanonical
        minDist <- input$minDist
        maxLogKd <- input$maxLogKd
        withProgress(message=msg, detail=detail, value=1, max=3, {
          res$hits = findSeedMatches(
              scantarget, scanmods, keepMatchSeq=keepmatchseq,
              minDist=minDist, maxLogKd=maxLogKd, shadow=shadow,
              onlyCanonical=onlyCanonical, p3.extra=TRUE, BP=BP )
        })
      }
      if(length(res$hits)>0){
        res$hits$log_kd <- (res$hits$log_kd/1000)
        res$hits <- res$hits[order(res$hits$log_kd)]
        names(res$hits) <- NULL
      }
      res$cs <- cs
      res$last <- res$time <- Sys.time()
      res$size <- object.size(res$hits)
      res$collection <- input$mirlist
      res$nsel <- nm <- length(selmods())
      res$sel <- ifelse(nm>1,paste(nm,"models"),input$mirnas)
      res$seq <- target()
      res$seqFeature <- input$seqFeature
      res$maxLogKd <- input$maxLogKd
      res$target_length <- nchar(target())
      if(input$subjet_type=="custom"){
        res$target <- "custom sequence"
      }else{
        res$target <- paste0(input$gene, " - ", seltx(),
                            " (", input$seqFeature, ")")
      }
      waiter_hide()
      return(res)
    })

    output$scan_target <- renderText({
      if(is.null(current.cs()) || is.null(cached.hits[[current.cs()]]))
        return(NULL)
      paste("Scan results in: ", cached.hits[[current.cs()]]$target)
    })

    output$cache.info <- renderText({
      if(cache.size()==0) return("Cache empty.")
      paste0(length(cached.checksums()), " results cached (",
             round(cache.size()/1024^2,3)," Mb)")
    })

    output$cached.results <- renderUI({
      ch <- cached.checksums()
      ch2 <- names(ch)
      names(ch2) <- vapply(ch, FUN.VALUE=character(1), FUN=function(x){
        paste0(x$time, ": ", x$sel, " on ", x$target, " (",
               format(x$size,units="Kb"),")")
      })
      radioButtons("selected.cache", "Cached results", choices=ch2)
    })

    observeEvent(input$loadCache, {
      if(is.null(input$selected.cache)) return(NULL)
      current.cs(input$selected.cache)
    })

    observeEvent(input$deleteCache, {
      if(is.null(input$selected.cache)) return(NULL)
      cached.hits[[input$selected.cache]] <- NULL
      if(current.cs()==input$selected.cache) current.cs(NULL)
    })

    output$hits_table <- renderDT({ # prints the current hits
      if(is.null(hits()$hits)) return(NULL)
      h <- as.data.frame(hits()$hits)
      h <- h[,setdiff(colnames(h), c("seqnames","width","strand") )]
      h <- tryCatch(h[order(h$log_kd),], error=function(e) return(h))
      dtwrapper(h, selection="single", callback=JS('
        table.on("dblclick.dt","tr", function() {
          Shiny.onInputChange("dblClickMatch", table.row(this).data()[0]+"/"+Math.random())
          var box = $("#box_match").closest(".box")
          if (box.hasClass("collapsed-box")){
            box.find("[data-widget=collapse]").click();
          }
        })
      '))
    })

    output$dl_hits <- downloadHandler(
      filename = function() {
        if(is.null(hits()$hits)) return(NULL)
        fn <- paste0("hits-", gsub("\\.[09]+", "",
                                   cached.hits[[current.cs()]]$target))
        if(hits()$nsel == 1){
          fn <- paste0(fn,"-",cached.hits[[cs]]$sel,".csv")
        }else{
          fn <- paste0(fn,"-",Sys.Date(),".csv")
        }
        fn
      },
      content = function(con) {
        if(is.null(hits()$hits)) return(NULL)
        h <- as.data.frame(hits()$hits)
        h <- h[,setdiff(colnames(h), c("seqnames","width","strand") )]
        write.csv(h, con, col.names=TRUE)
      }
    )

    observeEvent(input$colHelp, .getHelpModal("hitsCol"))
    observeEvent(input$stypeHelp, .getHelpModal("stypes"))
    observeEvent(input$stypeHelp2, .getHelpModal("stypes"))
    observeEvent(input$manhattanHelp, .getHelpModal("manhattan"))
    observeEvent(input$help_collections, .getHelpModal("collections"))
    observeEvent(input$help_aggregatedHits, .getHelpModal("aggregatedHits"))
    
    output$bartel2009 <- renderImage({
      list(src=system.file("docs", "Bartel2009_sites.png", package="scanMiRApp"),
           contentType = 'image/png', width=772, height=576,
           alt="miRNA sites types from Bartel, Cell 2009")
    })
    
    agghits_data <- reactive({
      if(is.null(hits()$hits)) return(NULL)
      h <- hits()$hits
      ag <- scanMiR::aggregateMatches(h, keepSiteInfo=TRUE)
      ag <- ag[order(ag$repression),]
      ag$transcript <- ag$repression <- NULL
      
      h <- GRanges(rep(as.factor(hits()$target),nrow(h)), IRanges(h$start, h$end))
      mcols(h) <- mcols(hits()$hits)
      ag <- scanMiR::aggregateMatches(h, keepSiteInfo=TRUE)
      ag <- ag[order(ag$repression),]
      ag$transcript <- ag$repression <- NULL
      ag
    })

    output$agghits_table <- renderDT({
      if(is.null(agghits_data())) return(NULL)
      h <- as.data.frame(agghits_data())
      dtwrapper(h)
    })
    
    output$dl_agghits <- downloadHandler(
      filename = function() {
        if(is.null(agghits_data())) return(NULL)
        fn <- paste0("agghits-", gsub("\\.[09]+", "",
                                   cached.hits[[current.cs()]]$target))
        if(hits()$nsel == 1){
          fn <- paste0(fn,"-",cached.hits[[cs]]$sel,".csv")
        }else{
          fn <- paste0(fn,"-",Sys.Date(),".csv")
        }
        fn
      },
      content = function(con) {
        if(is.null(agghits_data())) return(NULL)
        h <- as.data.frame(agghits_data())
        write.csv(h, con, col.names=TRUE)
      }
    )
    
    ## end scan hits and cache

    manhattan_data <- reactive({
      if(is.null(hits()$hits)) return(NULL)
      h <- hits()$hits
      if(length(h)==0) return(NULL)
      h <- h[order(h$log_kd)]
      if(!is.null(h$miRNA) && length(unique(h$miRNA))>input$manhattan_n){
        mirs <- as.character(head(unique(h$miRNA),input$manhattan_n))
        h <- h[h$miRNA %in% mirs]
      }
      if(length(h)==0) return(NULL)
      h
    })

    output$manhattan <- renderPlotly({
      if(is.null(h <- manhattan_data()))
        return(ggplotly(ggplot(), source="manhattan"))
      plotlyObserver$resume()
      sn <- as.character(seqnames(h)[1])
      meta <- metadata(h)
      h <- as.data.frame(h)
      if(!is.null(input$manhattan_ordinal) && input$manhattan_ordinal){
        h$position <- order(h$start)
        xlab <- "Position (ordinal)"
        xlim <- c(1,length(h))
      }else{
        h$position <- round(rowMeans(h[,2:3]))
        xlab <- "Position (nt) in sequence"
        xlim <- c(1,hits()$target_length)
      }
      ael <- list(x="position", y="-log_kd", type="type")
      if("sequence" %in% colnames(h)) ael$seq="sequence"
      if("miRNA" %in% colnames(h)) ael$colour="miRNA"
      p <- ggplot(h, do.call(aes_string, ael))
      ymax <- max(-h$log_kd)
      if(length(selmods())==1){
        mer8 <- get8merRange(selmods()[[1]])/-1000
        ymax <- max(mer8)
        p <- p + geom_rect(aes(colour=NULL),
          data=data.frame(type="8mer range", log_kd=0, position=1),
          xmin=xlim[1], xmax=xlim[2], ymin=min(mer8), ymax=max(mer8),
          alpha=0.2, fill="green")
      }
      p <- p + theme_minimal() + theme(axis.line.x=element_line()) +
        geom_hline(yintercept=-hits()$maxLogKd, linetype="dashed",
                   color="red", size=1) +
        geom_point(size=2) + xlab(xlab) + expand_limits(x=xlim, y=c(0,ymax))
      if(!is.null(h$ORF) && !is.null(meta$tx_info)){
        orflen <- meta$tx_info[sn, "ORF.length"]
        p <- p + geom_vline(xintercept=orflen, color="grey", size=1)
      }
      p <- p + scale_x_continuous(limits=c(0,xlim[2]), expand=c(0,0)) +
        scale_y_continuous(limits=c(0,ymax), expand=c(0,0.1))
      ggplotly(p, source="manhattan")
    })

    selectedMatch <- reactiveVal()

    observeEvent(input$dblClickMatch, {
      if(is.null(hits()$hits)) return(NULL)
      if(is.null(input$dblClickMatch)) return(NULL)
      rid <- as.integer(strsplit(input$dblClickMatch, "/", fixed=TRUE)[[1]][[1]])
      if(is.null(rid) || !(rid>0)) return(NULL)
      h <- hits()$hits
      selectedMatch(h[order(h$log_kd)[rid]])
      showModal(modalDialog(
        title = "Target alignment",
        textOutput("alignment_header"),
        verbatimTextOutput("alignment"),
        easyClose = TRUE,
        footer = NULL
      ))
    })

    plotlyObserver <- observeEvent(event_data("plotly_click", "manhattan",
                                          priority="event"), suspended=TRUE, {
      if(is.null(h <- manhattan_data())) return(NULL)
      event <- event_data("plotly_click", "manhattan")
      if(!is.list(event) || is.null(event$pointNumber)) return(NULL)
      rid <- as.integer(event$pointNumber+1)
      if(is.null(rid) || !(rid>0)) return(NULL)
      if(!is.null(h$miRNA) && length(unique(h$miRNA))>1 &&
         !is.null(event$curveNumber)){
        h <- h[as.integer(droplevels(h$miRNA))==as.integer(event$curveNumber)]
      }
      selectedMatch(h[rid])
      showModal(modalDialog(
        title = "Target alignment",
        textOutput("alignment_header"),
        verbatimTextOutput("alignment"),
        easyClose = TRUE,
        footer = NULL
      ))
    })

    output$alignment_header <- renderText({
      if(is.null(m <- selectedMatch()))
        return("Double-click on a row of the table above to visualize it here")
      miRNA <- ifelse("miRNA" %in% colnames(mcols(m)),
                      as.character(mcols(m)$miRNA),hits()$sel)
      paste0(miRNA, " match at ",start(m),"-",end(m)," (", mcols(m)$type, ")")
    })

    output$alignment <- renderPrint({
      if(is.null(m <- selectedMatch())) return(NULL)
      mir <- ifelse("miRNA" %in% colnames(mcols(m)),
                    as.character(m$miRNA), hits()$sel)
      mod <- modlists[[hits()$collection]][[as.character(mir)]]
      seqs <- hits()$seq
      seqs <- setNames(as.character(seqs), as.character(seqnames(m)))
      viewTargetAlignment(m, mod, seqs=seqs)
    })

    ##############################
    ## miRNA-centric tab

    mod <- reactive({ # the currently-selected KdModel
      if(is.null(allmods()) || is.null(input$mirna)) return(NULL)
      allmods()[[input$mirna]]
    })

    output$modconservation <- renderText({
      if(is.null(mod())) return(NULL)
      as.character(conservation(mod()))
    })

    output$mirbase_link <- renderUI({
      tags$a(href=paste0("http://www.mirbase.org/textsearch.shtml?q=",
                         input$mirna),
             icon("external-link"), "miRBase", target="_blank")
    })

    output$modplot <- renderPlot({ # affinity plot
      if(is.null(mod())) return(NULL)
      plotKdModel(mod())
    })

    output$targets_ui <- renderUI({
      if(is.null(annotations[[input$mirlist]]$aggregated)){
        return(tags$p("Targets not accessible ",
                      "(no pre-compiled scan available)"))
      }
      list(
        fluidRow(
          column(5, tags$p("Double-click on a row to visualize hits.")),
          column(4, checkboxInput("targetlist_gene",
                                  "Only top transcript per gene",value=FALSE)),
          column(3, actionButton("stypeHelp2","Site types",
                                 icon=icon("question-circle")))
        ),
        withSpinner(DTOutput("mirna_targets")),
        downloadLink('dl_mirTargets', label = "Download all")
      )
    })

    txs <- reactive({ # the tx to gene symbol table for the current annotation
      if(is.null(input$mirlist) || input$mirlist=="" ||
         !(input$mirlist %in% names(annotations))) return(NULL)
      db <- annotations[[input$mirlist]]$ensdb
      if(is.null(db)) return(NULL)
      if(is(db,"EnsDb")){
        tx <- mcols(transcripts(db, c("tx_id","gene_id","tx_biotype")))
        tx <- merge(tx,mcols(genes(db, c("gene_id","symbol"))), by="gene_id")
        tx <- as.data.frame(tx[,c("symbol","tx_id","tx_biotype")])
      }else{
        tx <- mcols(transcripts(db, c("gene_id","tx_name","TXTYPE")))
        colnames(tx) <- c("gene","tx_id","tx_biotype")
      }
      tx
    })

    mirtargets_prepared <- reactive({
      if(is.null(preTargets <- annotations[[input$mirlist]]$aggregated) ||
         !(input$mirna %in% names(preTargets))) return(NULL)
      d <- preTargets[[input$mirna]]
      d$repression <- d$repression/1000
      if(!is.null(annotations[[input$mirlist]]$addDBs)){
        for(f in names(annotations[[input$mirlist]]$addDBs)){
          x <- annotations[[input$mirlist]]$addDBs[[f]]
          x <- x[x$miRNA==input$mirna,]
          row.names(x) <- x$transcript
          d[[f]] <- x[as.character(d$transcript),"score"]
        }
      }
      if(!is.null(txs())){
        d <- merge(txs(), d, by.x="tx_id", by.y="transcript")
        colnames(d) <- gsub("tx_id", "transcript", colnames(d))
        if(input$targetlist_gene){
          d <- d[order(d$repression),]
          d <- d[!duplicated(d$symbol),]
        }
      }
      as.data.frame(d[order(d$repression),])
    })

    output$mirna_targets <- renderDT({
      d <- mirtargets_prepared()
      if(is.null(d)) return(NULL)
      colnames(d) <- gsub("^n\\.","",colnames(d))
      dtwrapper(d, selection="single", callback=JS('
      table.on("dblclick.dt","tr", function() {
        Shiny.onInputChange("dblClickSubject", table.row(this).data()[1])
      })
    '))
    })

    ## double-click on a transcript in miRNA targets:
    observeEvent(input$dblClickSubject, {
      sub <- input$dblClickSubject
      if(input$targetlist_gene) return(NULL)
      gene <- getGeneFromTx(sel_ensdb(), sub)
      isolate(updateSelectizeInput(session, "gene", selected=gene,
                           choices=allgenes(), server=TRUE))
      updateTabItems(session, "subject_type", "transcript")
      txs <- getTxs(sel_ensdb(), gene=gene)
      #updateCheckboxInput(session, "utr_only", value=input$targetlist_utronly)
      updateSelectizeInput(session, "transcript", selected=sub, choices=txs)
      updateSelectizeInput(session, "mirnas", selected=input$mirna)
      #updateTabItems(session, "main_tabs", "tab_subject")
      newflag <- changeFlag()+1
      changeFlag(newflag)
      observe({
        tmp <- changeFlag()
        cs <- checksum()
        cached.hits[[cs]] <- do.scan()
        current.cs(cs)
      })
      updateTabItems(session, "main_tabs", "tab_hits")
    })

    output$dl_mirTargets <- downloadHandler(
      filename = function() {
        if(is.null(input$mirna)) return(NULL)
        paste0(input$mirna, "targets.csv")
      },
      content = function(con) {
        write.csv(mirtargets_prepared(), con, col.names=TRUE)
      }
    )

    output$pkgVersions <- renderText({
      paste(
        "Running on scanMiR", packageVersion("scanMiR"), "and scanMiRApp",
        packageVersion("scanMiRApp")
      )
    })

    waiter_hide()
  }
}


#' scanMiRApp
#' A wrapper for launching the scanMiRApp shiny app
#'
#' @param annotations A named list of \code{\link{ScanMiRAnno}} objects. If
#' omitted, will use the base ones.
#' @param ... Passed to \code{\link{scanMiRserver}}
#'
#' @return A shiny app
#' @export
#' @examples
#' if(interactive()){
#'   anno <- ScanMiRAnno("fake")
#'   scanMiRApp(list(fakeAnno=anno))
#' }
scanMiRApp <- function(annotations=NULL, ...){
  if(is.null(annotations)){
    an <- c("GRCh38","GRCm38","Rnor_6")
    message("Loading annotations for ", paste(an, collapse=", "))
    annotations <- lapply(an, FUN=ScanMiRAnno)
  }
  shinyApp(scanMiRui(), scanMiRserver( annotations = annotations, ... ))
}

#' scanMiRui
#'
#' UI for the scanMiR app.
#'
#' @import shiny shinydashboard
#' @importFrom plotly plotlyOutput ggplotly
#' @importFrom shinycssloaders withSpinner
#' @importFrom rintrojs introjsUI introBox
#' @importFrom shinyjqui jqui_resizable
#' @importFrom waiter use_waiter waiter_show_on_load spin_1
#' @return A shiny ui
#' @export
#' @examples
#' ui <- scanMiRui()
scanMiRui <- function(){
  scanMiRlogo <- paste0("https://raw.git",
                        "hubusercontent.com/ETHZ-INS/scanMiR/",
                        "master/inst/docs/sticker.svg")
  mer12scheme <- paste0("https://raw.git",
                        "hubusercontent.com/ETHZ-INS/scanMiR/",
                        "master/inst/docs/12mer.png")
  dashboardPage( skin="black",

    dashboardHeader(title = "scanMiRApp", titleWidth = "300px",
                    tags$li(class="dropdown",
                            actionLink("helpBtn", label="Quick start",
                                       icon=icon("question")))),

    ## Sidebar content
    dashboardSidebar( width = "200px",
      sidebarMenu(id="main_tabs",
        menuItem("About", tabName="tab_about"),
        menuItemOutput("menuCollection"),
        menuItem("Search in\ngene/sequence", id="menuSearch",
          menuSubItem("subject", "tab_subject"),
          menuItemOutput("menuMirnas"),
          menuSubItem("options", "tab_options"),
          menuSubItem("hits", "tab_hits"),
          startExpanded=TRUE
        ),
        menuItem(tags$span(id="menuMiRNA", "miRNA-based"), tabName="tab_mirna")
      ),
      tags$a(
        href=paste0("https://git","hub.com/ETHZ-INS/scanMiR"), target="_blank",
        tags$img(src=scanMiRlogo),
        style="display: block; position: fixed; bottom: 5px; left: 20px;")
    ),
    ## Body Content
    dashboardBody(
      introjsUI(),
      use_waiter(),
      waiter_show_on_load(html = tagList(
        tags$img(src=scanMiRlogo), tags$br(), tags$br(),
        tags$h3("Please wait while the application is initialized..."),
        spin_1()
      )),
      tabItems(
        tabItem( tabName = "tab_collection",
          tags$h3("Select a miRNA collection"), tags$br(),
          column(6,
            tabBox(id="collection_type", width=12,
              tabPanel( title="Pre-built", value="prebuilt",
                        actionButton(inputId="help_collections", label="",
                          style="float:right;", icon=icon("question-circle")),
                     selectInput("mirlist", "miRNA collection", choices=c()))
            )),
          column(6, valueBoxOutput("selected_collection", width=12)),
          box(width=12, title="Extra collection information", 
              collapsible=TRUE, collapsed=TRUE,
              withSpinner(verbatimTextOutput("collection_summary")))
        ),
        tabItem(tabName="tab_subject",
          tags$div(id="subjectsbox", style="min-height: 200px;",
          tabBox(
            title="Search bindings in", side="right",
            id="subjet_type", width=12, selected = "transcript",
            tabPanel(
              title="Custom sequence", value="custom",
              textAreaInput(
                inputId="customseq", label="Sequence (RNA or DNA)",
                placeholder=paste0("Paste in here a sequence in which you want",
                                   " to search for binding sites"),
                width = "100%", height = "250px"),
              checkboxInput(inputId="circular", label="circularize"),
              verbatimTextOutput("custom_info"),
              actionButton("rndseq", "Generate random sequence")
            ),
            tabPanel(
              title="Transcript", value="transcript",
              selectizeInput("annotation", "Genome & Annotation", choices=c()),
              tags$div(id="txbox",
              tags$div(selectizeInput("gene", "Gene", choices=c()),
                       style="float: left; padding-right: 10px;"),
              tags$br(), tags$p(style="text-align: left;", "Type in the first ",
                            "few letters of a gene name to populate the list"),
              htmlOutput("gene_link"),
              tags$div(style="clear: left;"),
              selectizeInput("transcript", "Transcript", choices=c()),
              introBox(selectInput("seqFeature", "Part to scan", 
                                   choices=c("3' UTR only", "CDS+UTR", "whole transcript")),
                       data.hint="You can decide here whether you want to scan 
                       just the 3' UTR or also include the coding sequence,
                       or scan the whole transcript."),
              withSpinner(tableOutput("tx_overview")))
            )
          )
        )),
        tabItem(tabName="tab_mirnas",
          box(id="mirnasbox", width=12,
            column(12, selectizeInput("mirnas", choices=c(), multiple=TRUE,
                                label="Selected miRNAs: (type in to filter)")),
            column(8, actionButton("mirnas_vert",
                          "Select miRNAs conserved across vertebrates"), 
             tags$br(), actionButton("mirnas_mammals",
                          "Select miRNAs conserved across mammals"),
             tags$br(), actionButton("mirnas_confident",
                                     "Select confidently annotated miRNAs")
            ),
            column(4, "The conservation status of miRNAs is obtained from",
                   tags$a(href="http://www.targetscan.org/", target="_blank",
                          "TargetScan")),
            column(8,
              checkboxInput("mirnas_all", "Search for all miRNAs")
            ),
            column(4, actionButton("mirnas_clear",
                                   "Clear all selected miRNAs"))
          )
        ),
        tabItem(tabName="tab_options",
          box( width=6,
            sliderInput("minDist",
              label=paste("Minimum distance in NT between two binding sites",
                           "of the same miRNA"),
              min = 0, max = 20, value = 7, step = 1),
            sliderInput(inputId="shadow",
              label=paste("Ribosomal shadow at the beginning of the 3'UTR,",
                          "recommended is 15 NT"),
              min = 0, max = 20, value = 15, step = 1),
            sliderInput(inputId="maxLogKd", label = "Maximum log_kd to report",
                        min = -4, max = 0, value = -1, step = 0.25)
          ),
          box(width=6, checkboxInput("keepmatchseq", "Keep matched sequence"),
              checkboxInput("scanNonCanonical",
                            "Search also for non-canonical sites", value=TRUE))
        ),
        tabItem(tabName="tab_hits",
          fluidRow(
            column(2, tags$div(id="scanBtnBox", 
              style="display: inline-block; min-height: 30px; min-width: 100px;",
              uiOutput("scanBtn"), HTML("&nbsp;"))),
            column(10, tags$h5(textOutput("scan_target")))
          ),
          tags$div(id="hitsbox", style="min-height: 300px;",
          box(width=12, collapsible=TRUE, collapsed=TRUE,
              title="Display sites along transcript",
              actionButton("manhattanHelp", label="", style="float: right;", 
                           icon=icon("question-circle")),
              tags$p("Hover on points to view details, and click to ",
                      "visualize the alignment on the target sequence. You may
                      also select miRNAs to show/hide by clicking on the legend."),
              withSpinner(plotlyOutput("manhattan")),
            column(6, numericInput("manhattan_n", "Number of top miRNAs displayed",
                                   value=10, min=1, max=50)),
            column(6, checkboxInput("manhattan_ordinal", "Ordinal position",
                                    value=FALSE))
          ),
          box(width=12, title="Aggregated per miRNA", collapsible=TRUE, collapsed=TRUE,
              withSpinner(DTOutput("agghits_table")),
              column(6,tags$p(downloadLink('dl_agghits', label = "Download all"))),
              column(6, actionButton("help_aggregatedHits","About aggregation...",
                                     icon=icon("question-circle")))
          ),
          box(width=12, title="Table", collapsible=TRUE,
            fluidRow(
              column(6,
                tags$p("Double click on a row to visualize the alignment",
                        " on the target sequence.")),
              column(6, style="text-align: right;", 
                   actionButton("colHelp","What are those columns?",
                                icon=icon("question-circle")),
                   actionButton("stypeHelp","Site types",
                                icon=icon("question-circle"))
              )
            ),
            withSpinner(DTOutput("hits_table")),
            tags$p(downloadLink('dl_hits', label = "Download all"))
            
          )),
          box(width=12, title="Cached hits", collapsible=TRUE, collapsed=TRUE,
            textOutput("cache.info"),
            uiOutput("cached.results"),
            actionButton("loadCache", "Load"),
            actionButton("deleteCache", "Delete")
          )
        ),

        # miRNA-based
        tabItem(tabName="tab_mirna",
          fluidRow(
            column(5, tags$div(id="mirnainput",
                        selectizeInput("mirna", "miRNA", choices=c()))),
            column(4, tags$strong("TargetScan conservation status"), 
                  textOutput("modconservation")),
            column(3, htmlOutput("mirbase_link"))
          ),
          tags$div(id="affinitybox", style="min-height: 40px;",
            box(width=12, title="Affinity plot", collapsible=TRUE, 
                collapsed=TRUE,
                withSpinner(jqui_resizable(plotOutput("modplot")))
            )
          ),
          tags$div(id="targetsbox", style="min-height: 100px;",
            box(width=12, title="Targets", collapsible=TRUE, 
                uiOutput("targets_ui"))
          )
        ),
        tabItem(tabName = "tab_about",
## TAB ABOUT
    box(width=12, title="miRNA binding prediction",
      tags$div(style="font-size: 110%;",
        tags$img(src=mer12scheme, style="float: right; margin: 20px; width: 25%;"), 
        tags$p("micro-RNAs (miRNAs) regulate transcripts stability and 
          translation by targeting the Argonaute complex to RNAs exhibiting a 
          partial complementarity with, in particular, the miRNA's seed 
          sequence. For a general review on miRNA biology and targeting, see
               ", tags$a(href="https://doi.org/10.1016/j.cell.2018.03.006", 
                         "Bartel (2018)",.noWS = "outside"),".", 
               tags$a(href="https://dx.doi.org/10.1126/science.aav1741", 
            "McGeary, Lin et al. (2019)"), " have additionally shown the 
      relevance of flanking nucleotides by empirically measuring the affinity 
      (specifically the dissociation rate constant, KD) of a set of miRNAs to 
      random 12-mer sequences, and then computationally extrapolating to other 
      miRNAs). The ",
      tags$a( href=paste0("https://git","hub.com/ETHZ-INS/scanMiR"),
              target="_blank", "scanMiR package",.noWS = "outside"), ", to which this app is an 
      interface, builds on this work to offer a powerful and flexible binding 
      and repression prediction framework."))
    ),
    box(width=12, title="Getting started",
        tags$div( style="font-size: 110%;",
           tags$p("For a quick tour of the app, ", actionLink("helpLink", "click here"),"."),
           tags$p("There are two main ways to use scanMiRApp:"),
           tags$br(), tags$h4("Transcript-centered:"),
           tags$p("In the 'Search in gene/sequence' menu, you'll be able to ",
                  "scan the sequence of a given transcript for binding sites ",
                  "of (sets of) miRNA(s)."),
           tags$br(), tags$h4("miRNA-centered:"),
           tags$p("In the 'miRNA-based' tab on the left, you'll be able to ",
                  "visualize information relative to a selected miRNA, ",
                  "including for instance its general binding profile and its",
                  "top targets.")
        )
    ),
    tags$div(style="text-align: right; font-weight: normal;",
      tags$p("The shiny app was developed by Pierre-Luc Germain and ",
           "Michael Soutschek in the context of broader research in the ",
           tags$a( href="http://schrattlab.ethz.ch", "Schratt lab",
                   target="_blank",.noWS = "outside"), ".",
           tags$br(), "Bugs reports and feature requests are welcome on the ",
           tags$a( href=paste0("https://git","hub.com/ETHZ-INS/scanMiRApp/issues"),
                   target="_blank", "github repository",.noWS = "outside"), "."),
           textOutput("pkgVersions")
      )
## END TAB ABOUT
        )
      ),
      tags$head(tags$style(HTML('.content-wrapper { overflow: auto; }')))
    )
  )
}

#' IndexedFst
#'
#' Objects of the IndexedFst class enable fast named random access to FST files.
#' This is particularly appropriate for large data.frames which often need to
#'  be accessed according to the (e.g. factor) value of a particular column.
#'
#' @examples
#' # we first create and save an indexed FST file
#' tmp <- tempdir()
#' f <- system.file(tmp, "test")
#' d <- data.frame( category=sample(LETTERS[1:4], 10000, replace=TRUE),
#'                  var2=sample(LETTERS, 10000, replace=TRUE),
#'                  var3=runif(10000) )
#' format(object.size(d),units="Kb")
#' saveIndexedFst(d, "category", f)
#' rm(d)
#' # we then load the index, and can use category names for random access:
#' d <- loadIndexedFst(f)
#' format(object.size(d),units="Kb")
#' nrow(d)
#' names(d)
#' head(d$A)
#' @seealso \code{\link{saveIndexedFst}}, \code{\link{loadIndexedFst}}
#' @export
#' @import methods
#' @exportClass IndexedFst
#' @author Pierre-Luc Germain, \email{pierre-luc.germain@@hest.ethz.ch}
#' @aliases IndexedFst-class IndexedFst
#' @return Depends on the method
#' @rdname IndexedFst-class
#' @name IndexedFst-class
setClass(
  "IndexedFst",
  slots=representation(
    fst.file="character",
    index="data.frame",
    add.info="list",
    nthreads="integer"
  ),
  prototype=prototype(fst.file=NA_character_, index=data.frame(), nthreads=1L,
                      add.info=list()),
  validity=function(object){
    if(length(object@nthreads)!=1 || !is.integer(object@nthreads) ||
       object@nthreads<1)
      stop("`nthreads` should be a positive integer")
    if(length(object@fst.file)!=1)
      stop("fst.file should be a character of length 1.")
    if(!file.exists(object@fst.file)) stop("FST file does not exist!")
    TRUE
  }
)

setMethod("initialize", "IndexedFst", function(.Object, ...) {
  o <- callNextMethod(.Object, ...)
  o@fst.file <- normalizePath(o@fst.file)
  o@nthreads <- as.integer(o@nthreads)
  ff <- gsub("\\.fst$",".idx.rds",o@fst.file)
  tryCatch({
    ff <- readRDS(ff)
    o@index <- ff$index
    ff$index <- NULL
    o@add.info <- ff
  }, error=function(e) stop("Could not find or read index file."))
  validObject(o)
  return(o)
})


#' @rdname IndexedFst-class
#' @importMethodsFrom methods show
#' @param object an IndexedFst object
#' @export
setMethod("show", "IndexedFst", function(object){
  paste0(object@fst.file, " (",nrow(object@index)," sets)")
})

#' @rdname IndexedFst-class
#' @importFrom fst fst.metadata
#' @export
setMethod("summary", "IndexedFst", function(object){
  fst.metadata(object@fst.file)
})

#' @rdname IndexedFst-class
#' @param x an IndexedFst object
#' @export
setMethod("names", "IndexedFst", function(x){
  row.names(x@index)
})

#' @rdname IndexedFst-class
#' @export
setMethod("length", "IndexedFst", function(x){
  nrow(x@index)
})

#' @rdname IndexedFst-class
#' @export
setMethod("lengths", "IndexedFst", function(x){
  y <- x@index[,2]-x@index[,1]+1
  names(y) <- names(x)
  y
})

setGeneric("nrow")
#' @rdname IndexedFst-class
#' @export
setMethod("nrow", "IndexedFst", function(x){
  max(x@index[,2])
})

setGeneric("ncol")
#' @rdname IndexedFst-class
#' @export
setMethod("ncol", "IndexedFst", function(x){
  length(fst.metadata(x@fst.file)$columnNames)
})

setGeneric("colnames")
#' @rdname IndexedFst-class
#' @export
setMethod("colnames", signature("IndexedFst"), function(x){
  fst.metadata(x@fst.file)$columnNames
})

#' @rdname IndexedFst-class
#' @param i the desired index (either numeric or name)
#' @param j,drop ignored
#' @param ... ignored
#' @export
setMethod("[[", signature("IndexedFst"), function(x, i, j=NULL, ...){
  if(is.numeric(i)){
    name <- names(x)[i]
  }else{
    name <- i
  }
  .fst.read.wrapper(x, name)
})

#' @rdname IndexedFst-class
#' @export
setMethod("[", signature("IndexedFst"), function(x, i, j=NULL, ..., drop=TRUE){
  if(is.logical(i)) i <- which(i)
  if(is.numeric(i)){
    name <- names(x)[i]
  }else{
    name <- i
  }
  .fst.read.wrapper(x, name)
})

#' @rdname IndexedFst-class
#' @param name the indexed name to fetch
#' @export
setMethod("$", "IndexedFst", definition = function(x, name){
  .fst.read.wrapper(x, match.arg(name, row.names(x@index)))
})

#' @rdname IndexedFst-class
#' @param n the desired number of rows
#' @export
setMethod("head", "IndexedFst", definition = function(x, n=6L, ...){
  if(!is.numeric(n) || !(n>0) || n!=as.integer(n))
    stop("`n` should be a positive integer.")
  w <- which(x@index[,2]>=n)
  if(length(w)==0) return(.fst.read.wrapper(x))
  head(x[seq_len(w[1])], n=n)
})

#' @rdname IndexedFst-class
#' @export
setMethod("as.data.frame", "IndexedFst", definition=function(x, name) {
  .fst.read.wrapper(x, convertGR=FALSE)
})

#' @importFrom fst threads_fst read.fst
#' @importFrom data.table rbindlist
#' @importFrom IRanges IRanges
#' @importFrom GenomicRanges GRanges
.fst.read.wrapper <- function(x, names=NULL, convertGR=TRUE){
  if(!is(x,"IndexedFst")) stop("`x` should be an IndexedFst object.")
  if(!is.null(names) && length(names)>1){
    f <- lapply(names, FUN=function(i) .fst.read.wrapper(x,i,convertGR=FALSE))
    if(length(f)>5){
      f <- as.data.frame(data.table::rbindlist(f))
    }else{
      f <- do.call(rbind, f)
    }
  }else{
    ont <- threads_fst()
    threads_fst(x@nthreads)
    if(is.null(names)){
      f <- read.fst(x@fst.file)
    }else{
      f <- read.fst(x@fst.file, from=x@index[names,1], to=x@index[names,2])
    }
    threads_fst(ont)
  }
  if(convertGR && !is.null(x@add.info$isGR) && x@add.info$isGR){
    f <- tryCatch(.df2gr(f, x@add.info), error=function(e){
      warning("Could not convert to GRanges:", e)
      f
    })
  }
  f
}

#' Saving and loading IndexedFst
#'
#' Functions to save or load and indexed \code{\link[fst]{fst}} file
#'
#' @param file Path to the fst file, it's index (.idx), or their prefix.
#' @param nthreads Number of threads to use for reading (default 1). This does
#' not affect the loading of the index itself, but will affect all downstream
#' reading operations performed on the object. If NULL, will use
#' `fst::threads_fst()`.
#'
#' @return `loadIndexedFst` returns an object of class
#' \code{\link{IndexedFst-class}}, and `saveIndexedFst` returns nothing.
#' @seealso \code{\link{IndexedFst-class}}
#' @rdname save-load-IndexedFst
#' @aliases loadIndexedFst
#' @export
#' @examples
#' # we first create and save an indexed FST file
#' tmp <- tempdir()
#' f <- system.file(tmp, "test")
#' d <- data.frame( category=sample(LETTERS[1:4], 10000, replace=TRUE),
#'                  var2=sample(LETTERS, 10000, replace=TRUE),
#'                  var3=runif(10000) )
#' saveIndexedFst(d, "category", f)
#' # we then load the index, and can use category names for random access:
#' d <- loadIndexedFst(f)
loadIndexedFst <- function(file, nthreads=1){
  if(grepl("\\.fst$",file)) return(new("IndexedFst", fst.file=file))
  if(grepl("\\.idx\\.rds$",file))
    return(new("IndexedFst", fst.file=gsub("idx\\.rds$","fst",file)))
  new("IndexedFst", fst.file=paste0(file,".fst"), nthreads=as.integer(nthreads))
}

#' saveIndexedFst
#'
#' Saves a data.frame (or GRanges object) into an indexed FST file.
#'
#' @param d A data.frame or \code{\link[GenomicRanges]{GRanges}} object
#' @param index.by A column of `d` by which it should be indexed.
#' @param file.prefix Path and prefix of the output files.
#' @param index.properties An optional data.frame of properties, with the levels
#' of `index.by` as row names.
#' @param add.info An optional list of additional information to save.
#' @param ... Passed to `write.fst`
#' @rdname save-load-IndexedFst
#' @aliases saveIndexedFst
#' @seealso \code{\link{IndexedFst-class}}
#' @importFrom fst write.fst threads_fst
#' @export
saveIndexedFst <- function(d, index.by, file.prefix, nthreads=1,
                           index.properties=NULL, add.info=list(), ...){
  if(is(d, "GRanges")){
    add.info <- c(add.info, metadata(d))
    d <- as.data.frame(d)
    d$end <- NULL
    if(all(d$strand=="*")){
      d$strand <- NULL
    }else if(length(unique(d$strand))==1){
      add.info$strand <- as.factor(d$strand[1])
      d$strand <- NULL
    }
    if(length(unique(d$width))==1){
      add.info$width <- d$width[1]
      d$width <- NULL
    }
    if(length(unique(d$seqnames))==1){
      add.info$seqnames <- d$seqnames[1]
      d$seqnames <- NULL
    }
    add.info$isGR <- TRUE
  }
  if(!is.data.frame(d)) stop("`d` should be a data.frame.")
  if((!is.character(index.by) && !is.integer(index.by)) ||
     length(index.by)!=1 || is.null(d[[index.by]]))
    stop("`index.by` should be a scalar character or integer indicating the ",
         "column by which to index.")
  if(is.numeric(d[[index.by]]))
    warning("The `index.by` column selected is numeric!\n",
            "If the column is really continuous, rather than having many ",
            "repeated values, this might lead to very suboptimal behaviors!")
  if(!is.null(nthreads)) threads_fst(nthreads)
  d <- d[order(d[[index.by]]),]
  file.prefix <- gsub("\\.fst$","",file.prefix)
  w <- which(!duplicated(d[[index.by]]))
  idx <- data.frame(row.names=d[[index.by]][w], start=w,
                    end=c(w[-1]-1L,nrow(d)))
  if(!is.null(index.properties)){
    idx <- merge(idx, as.data.frame(index.properties),
                 by="row.names", all.x=TRUE)
    row.names(idx) <- idx[,1]
    idx <- idx[,-1]
  }
  idx <- c(list(index=idx), add.info)
  saveRDS(idx, paste0(file.prefix, ".idx.rds"))
  write.fst(d, paste0(file.prefix, ".fst"), ...)
}

#' @importFrom GenomicRanges makeGRangesFromDataFrame
#' @importFrom S4Vectors metadata metadata<-
.df2gr <- function(x, add.info=list()){
  if(is.null(x$end)){
    if(is.null(x$width)){
      stopifnot(!is.null(add.info$width))
      x$end <- x$start + add.info$width-1L
    }else{
      x$end <- x$start + x$width-1L
    }
  }
  x$width <- NULL
  if(is.null(x$strand) && !is.null(add.info$strand))
    x$strand <- add.info$strand
  if(is.null(x$seqnames) && !is.null(add.info$seqnames))
    x$seqnames <- add.info$seqnames
  x <- makeGRangesFromDataFrame(x, keep.extra.columns = TRUE)
  add.info <- add.info[setdiff(names(add.info),
                               c("strand","width","seqnames","isGR"))]
  metadata(x) <- add.info
  x
}

anno <- ScanMiRAnno("fake")

test_that("runFullScan works and gives the expected output", {
  m1 <- runFullScan(anno, onlyCanonical=TRUE)
  m2 <- runFullScan(anno, extract="withORF", onlyCanonical=FALSE, maxLogKd=-0.5)
  expect_equal(seqlevels(m1), "ENSTFAKE0000056456")
  expect_equal(seqlevels(m2), "ENSTFAKE0000056456")
  expect_equal(table(m1$type)[["8mer"]],1L)
  expect_equal(table(m2$type)[["8mer"]],1L)
  expect_equal(table(m2$type)[["6mer-m8"]],1L)
  expect_gt(table(m2$type)[["non-canonical"]],0)
  expect_equal(table(m1$type)[["non-canonical"]],0L)
  expect_equal(length(unique(m2$ORF)),2L)
})

test_that("indexedFst works and maintains data integrity", {
  f <- file.path(tempdir(), "ifst.test")
  d <- data.frame( category=sample(LETTERS[1:4], 10000, replace=TRUE),
                   var2=sample(LETTERS, 10000, replace=TRUE),
                   var3=runif(10000) )
  saveIndexedFst(d, "category", f)
  d <- d[order(d$category),]
  d2 <- loadIndexedFst(f)
  d3 <- as.data.frame(d2)
  expect_equal(d$category, d3$category)
  expect_equal(d$var2, d3$var2)
  expect_equal(d$var3, d3$var3)
  expect_equal(nrow(d[d$category=="A",]), nrow(d2$A))
})

test_that("Pair enrichment analysis works", {
  eg <- expand.grid(target=LETTERS[2:5], mir=letters[1:5])
  i <- rpois(nrow(eg),2)
  i[i>7] <- 7
  eg <- eg[rep(seq_len(nrow(eg)), i),]
  eg <- rbind(eg, cbind(target=rep("A",12),mir=rep("a",12)))
  eg$type <- sample(c("8mer","7mer"), nrow(eg), replace=TRUE)
  eg$start <- 1L+sample.int(100,nrow(eg))
  gr <- GRanges(eg$target, IRanges(eg$start, eg$start+8), miRNA=eg$mir,
                type=factor(eg$type) )
  ep <- enrichedMirTxPairs(gr, max.binom.p=1, minSites=1)
  expect(ep[1,1]=="A" && ep[1,2]=="a", failure_message=
    "enrichedMirTxPairs does not report spiked result as top")
})

test_that("UTR plot works", {
  plotSitesOnUTR(tx="ENSTFAKE0000056456", annotation=anno,
                 miRNA="hsa-miR-155-5p")
  plotSitesOnUTR(tx="ENSTFAKE0000056456", annotation=anno,
                 anno$models[[1]]$canonical.seed)
})

#' enrichedMirTxPairs
#'
#' Identifies pairs of miRNA and target transcripts that have an unexpectedly
#' high number of sites.
#'
#' @param m A GRanges of matches, as produced by \code{\link{findSeedMatches}}.
#' This will be filtered down to only 8mer and 7mer sites.
#' @param minSites The minimum number of sites for a given miRNA-target pair to
#' be considered.
#' @param max.binom.p The maximum binomial p-value of miRNA-target pairs.
#'
#' @return A data.frame of top combinations, including number of sites and
#' the log-transformed binomial p-value.
#' @export
#'
#' @import Matrix
#' @importFrom stats pbinom
#' @examples
#' # we create a dummy scan (see `runFullScan`)
#' library(scanMiR)
#' seqs <- getRandomSeq(n=10)
#' mirs <- c("TTGTATAA","AGCATTAA")
#' m <- findSeedMatches(seqs,mirs,verbose=FALSE)
#' # we look for enriched pairs
#' res <- enrichedMirTxPairs(m, minSites=1, max.binom.p=1)
#' res
enrichedMirTxPairs <- function(m, minSites=5, max.binom.p=0.001){
  m <- m[as.integer(m$type) %in% grep("8mer|7mer",levels(m$type))]
  b <- .matches2sparse(m)
  b <- b[rowSums(b>=minSites)>0,,drop=FALSE]
  rs <- rowSums(b)
  cs <- colSums(b)
  p <- as.matrix(rs/sum(rs)) %*% t(cs/sum(cs))
  S <- pbinom(as.matrix(b)-1, prob=p, size=rs, lower.tail=FALSE, log.p=TRUE)
  mode(S) <- "integer"
  S <- as(round(S), "sparseMatrix")
  S <- .sparse2df(S, "logp.binom")
  b <- .sparse2df(b, "sites")
  S <- merge(S,b,by=c("feature","set"))
  rm(b)
  S <- S[S$logp.binom < as.integer(log(max.binom.p)) &
           S$sites>=as.integer(minSites),]
  S[order(S$logp.binom),]
}


#' @import Matrix
.matches2sparse <- function(x){
  as(as.matrix(table(as.factor(seqnames(x)), as.factor(x$miRNA))),
     "sparseMatrix")
}
.sparse2df <- function(x, content="value"){
  dimn <- dimnames(x)
  x <- summary(x)
  w <- which(x$x!=ifelse(is.logical(x$x),FALSE,0))
  xout <- data.frame( feature=factor(x$i[w], levels=seq_len(length(dimn[[1]])),
                                     labels=dimn[[1]]),
                      set=factor(x$j[w], levels=seq_len(length(dimn[[2]])),
                                 labels=dimn[[2]]) )
  if(!is.logical(x$x)) xout[[content]] <- x$x[w]
  xout
}

#' getTranscriptSequence
#'
#' Utility wrapper to extracts the sequence of a given transcript (UTR or
#' CDS+UTR).
#'
#' @param tx The ensembl ID of the transcript(s)
#' @param annotation A \code{\link{ScanMiRAnno}} object.
#' @param annoFilter An optional `AnnotationFilter` or `AnnotationFilterList` to
#' further filter the set of transcripts to be extracted
#' @param extract Which parts of the transcripts to extract. For `UTRonly`
#' (default) only the 3' UTR regions are extracted, `withORF` additionally
#' extracts the coding regions, and `exons` extracts all exons
#' @param ... Passed to \code{\link{AnnotationHub}}
#'
#' @return A \code{\link{DNAStringSet}}.
#' @export
#'
#' @importFrom GenomicFeatures exonsBy extractTranscriptSeqs cdsBy
#' threeUTRsByTranscript seqlevels<-
#' @importFrom GenomeInfoDb seqlevels
#' @importFrom ensembldb metadata seqlevelsStyle seqlevelsStyle<-
#' @importFrom AnnotationFilter AnnotationFilterList AnnotationFilter
#' @importFrom stats setNames
#' @import Biostrings
#' @examples
#' anno <- ScanMiRAnno("fake")
#' seq <- getTranscriptSequence( tx="ENSTFAKE0000056456", annotation=anno )
getTranscriptSequence <- function(tx=NULL, annotation, annoFilter=NULL,
                                  extract=c("UTRonly", "withORF", "exons"),...){
  if(!is.null(tx) && is(annotation$ensdb,"EnsDb"))
    tx <- gsub("\\.[0-9]+$","",as.character(tx))
  if(!is.null(annoFilter)){
    if(!is(annotation$ensdb,"EnsDb")){
      warning("`annoFilter` is ignored when the annotation is not in ",
              "`EnsDb` format")
      annoFilter <- NULL
    }else{
      if(!is(annoFilter, "AnnotationFilterList") &&
         !is(annoFilter, "AnnotationFilter"))
        stop("filter must be either `AnnotationFilter` or `AnnotationFilterList`")
      if(!is.null(tx))
        annoFilter <- AnnotationFilterList(annoFilter, ~tx_id %in% tx)
    }
  }
  if(is.null(annoFilter)){
    if(!is.null(tx)){
      annoFilter <- AnnotationFilter(~tx_id %in% tx)
    } else {
      annoFilter <- AnnotationFilterList()
    }
  }
  ensdb <- annotation$ensdb
  extract <- match.arg(extract)
  if(is(ensdb,"EnsDb")){
    if(extract=="exons") {
      gr <- exonsBy(ensdb, by="tx", filter=annoFilter)
    } else {
      gr <- suppressWarnings(threeUTRsByTranscript(ensdb, filter=annoFilter))
    }
  }else{
    if(is.null(tx)){
      if(extract=="exons") {
        gr <- exonsBy(ensdb, by="tx", use.names=TRUE)
      } else {
        gr <- suppressWarnings(threeUTRsByTranscript(ensdb, use.names=TRUE))
      }
    }else{
      if(extract=="exons") {
        gr <- .getExonsFromTxDb(tx, ensdb)
      }else{
        gr <- .get3UTRFromTxDb(tx, ensdb)
      }
    }
  }
  tx_strand <- unlist(unique(strand(gr)))
  genome <- annotation$genome
  if(is(ensdb,"EnsDb")) seqlevelsStyle(genome) <- "Ensembl"
  seqs <- DNAStringSet()
  if(length(gr)==0){
    if(extract=="withORF"){
      seqs <- DNAStringSet()
    } else {
      message("Nothing found!")
      return(DNAStringSet())
    }
  } else {
    gr <- gr[seqnames(gr) %in% seqlevels(genome)]
    seqs <- extractTranscriptSeqs(genome, gr)
  }
  if(extract=="withORF"){
    if(is(ensdb,"EnsDb")){
      grl_ORF <- suppressWarnings(cdsBy(annotation$ensdb, by="tx",
                                        filter=annoFilter))
    }else{
      if(is.null(tx)){
        grl_ORF <- suppressWarnings(cdsBy(annotation$ensdb, by="tx",
                                          use.names=TRUE))
      }else{
        grl_ORF <- .getExonsFromTxDb(tx, ensdb, fn=GenomicFeatures::cds)
      }
    }
    tx_strand <- unlist(unique(strand(grl_ORF)))
    if(length(grl_ORF)==0) {
      if(length(seqs) ==0) message("Nothing found!")
      return(seqs)
    }
    seqs_ORF <- extractTranscriptSeqs(genome, grl_ORF)
    orf.len <- setNames(lengths(seqs_ORF), names(seqs_ORF))
    seqs_ORF[names(seqs)] <- xscat(seqs_ORF[names(seqs)],seqs)
    seqs <- seqs_ORF
    rm(seqs_ORF)
    mcols(seqs)$ORF.length <- orf.len[names(seqs)]
  }
  mcols(seqs)$strand <- tx_strand[names(seqs)]
  seqs
}


#' @importFrom GenomicFeatures exons cds
.getExonsFromTxDb <- function(tx, db, fn=GenomicFeatures::exons){
  a <- fn(db,
          filter = list(tx_name = tx),
          columns = list("tx_name"))
  a$tx_name <- a$tx_name[which(a$tx_name %in% tx)]
  names(tx) <- tx <- unique(unlist(a$tx_name,use.names=FALSE))
  a <- GRangesList(lapply(tx, FUN=function(x){
    y <- a[any(a$tx_name==x)]
    y$tx_name <- x
    sort(y, decreasing=as.character(unlist(strand(y)))[1]=="-")
  }))
  names(a) <- tx
  a
}

.get3UTRFromTxDb <- function(tx, db){
  te <- .getExonsFromTxDb(tx, db, fn=GenomicFeatures::exons)
  if(length(te)==0) return(NULL)
  tcds <- .getExonsFromTxDb(tx, db, fn=GenomicFeatures::cds)
  i <- intersect(names(tcds[lengths(tcds)>0]), names(te[lengths(te)>0]))
  if(length(i)==0) return(NULL)
  te <- GRangesList(mapply(te=te[i], tcds=tcds[i], FUN=function(te,tcds){
    te <- setdiff(te,tcds)
    if(as.character(unlist(strand(te),use.names=FALSE)[1])=="-"){
      return(te[end(te)<=min(start(tcds))])
    }
    te[start(te)>=max(end(tcds))]
  }))
  te
}


#' plotSitesOnUTR
#'
#' Wrapper function with minimal arguments to plot scanMiR-Binding
#' sites on 3'UTRs of specified transcripts. The red dashed line indicates the
#' background threshhold is indicated, the lightblue dashed line shows the
#' average 8mer dissociation rate of the given miRNA
#'
#' @param tx An ensembl TranscriptID
#' @param annotation A \code{\link{ScanMiRAnno}} object.
#' @param miRNA A miRNA name in the mirbase format (eg. "hsa-miR-485-5p"), a
#' `KdModel`, or a miRNA sequence or target seed.
#' @param label_6mers Logical whether to label 6mer sites in the plot
#' @param label_notes Logical whether to label special sites in the plot (as
#'   TDMD or Slicing)
#' @param verbose Logical; whether to print updates on the processing
#' @param ... Any further arguments passed to
#' \code{\link[scanMiR]{findSeedMatches}}
#'
#' @return Returns a ggplot.
#' @importFrom ggplot2 ggplot geom_hline geom_point geom_text labs xlim aes
#' theme_light
#' @import scanMiR
#' @export
#' @examples
#' anno <- ScanMiRAnno("fake")
#' plotSitesOnUTR( tx="ENSTFAKE0000056456", annotation=anno,
#'                 miRNA="hsa-miR-155-5p" )
plotSitesOnUTR <- function(tx, annotation, miRNA = NULL, label_6mers = FALSE,
                           label_notes = FALSE, verbose = TRUE, ...){
  stopifnot(is(annotation, "ScanMiRAnno"))
  if(verbose) message("Prepare miRNA model")
  mods <- annotation$models
  if((is(miRNA, "character") && length(miRNA) == 1 &&
     nchar(gsub("[ACGTU]", "", miRNA)) == 0) || is(miRNA, "KdModel")){
    mods <- miRNA
    if(!(is(miRNA,"KdModel"))){
      title <- paste0(tx, " - ", miRNA)
    }
  }else if(miRNA %in% names(mods)){
    mods <- mods[[miRNA]]
  }else if(length(w <- grep(miRNA, names(mods), ignore.case=TRUE)) == 1){
    mods <- mods[[w]]
  }else{
    stop("The specified microRNA is not listed for this species. Please check",
         "\nthe spelling (eg. 'hsa-mir-485-5p') and the organism")
  }
  if(verbose)
    message("Get Transcript Sequence")
  Seq <- getTranscriptSequence(tx = tx, annotation = annotation)
  if(verbose)
    message("Scan")
  m <- findSeedMatches(seqs = Seq, seeds = mods, shadow = 15L,
                       keepMatchSeq = TRUE, p3.extra = TRUE, ret = "data.frame",
                       verbose = FALSE, ...)
  if(isFALSE(label_6mers)){
    m$type <- ifelse(grepl("6mer", m$type), "", as.character(m$type))}
  if(is(mods, "KdModel")){
    m$logKd <- m$log_kd/1000
    m$type <- ifelse(m$type == "non-canonical", "", as.character(m$type))
    m <- as.data.frame(m[m$logKd < -1, ])
    mer8 <- get8merRange(mods)/-1000
    max8 <- max(mer8)
    title <- paste0(tx, " - ",mods$name)
    p <- ggplot(m, aes(x = start, y = -logKd)) +
      geom_hline(yintercept = 1,linetype = "dashed", color = "red", size = 1) +
      geom_point(size = 2) + geom_text(label = m$type, nudge_y = -max8/50) +
      labs(x = "sequence length", y = "-logKd", title = title) +
      theme_light() + xlim(0, width(Seq)) + expand_limits(y=c(0,max8)) +
      geom_rect(
               xmin = -Inf, xmax = Inf,
               ymin = min(mer8), ymax = max8,  fill = "chartreuse3", alpha=.3)
    if(label_notes && !("note" %in% colnames(m))){
      label_notes <- FALSE
      if(verbose) message("No label_notes to plot")
    }
    if(label_notes){
      p <- p + geom_text(data = m[m$note != "-", ], label = aes(note),
                         nudge_y = max8/50)
    }
  }else{
    m <- m[m$type != "",]
    p <- ggplot(m, aes(x = start, y = type)) +
      geom_point(size = 2) +
      labs(x = "sequence length", y = "site type", title = title) + xlim(0, width(Seq)) +
      theme_light()
  }
  p
}

.getAppIntro <- function(){
  data.frame(
    element=c(NA, "#menuCollection", "#menuSearch", "#menuMiRNA", "#mirnainput",
              "#affinitybox", "#targetsbox", NA, "#subjectsbox", "#txbox",
              "#mirnasbox", "#scanBtn", "#hitsbox", NA),
    intro=c(
      "This introduction will walk you through the usage of the scanMiR app.",
      "First, use this tab to select the miRNA collection relative to your 
      species of interest",
      "In this section, you can scan and visualize specific transcripts or 
      sequence for miRNA binding sites",
      "In this section, you can instead start from a specific miRNA, visualize 
      its binding profile, and get its top predicted targets.<br/><br/>
      We'll start by looking at that part.",
      "Use this dropdown to select the miRNA you're interested in.<br/><br/>
      Note that you don't have to scroll down until you find it - you can 
      simply erase what's written in the box, type the beginning of the miRNA 
      name, and see the matching options.",
      "This box contains a plot summarizing the binding profile of the miRNA, 
      plotting the dissociation rate (i.e. affinity) of the top 7mers sequences
      (with or without the 'A' at position 1).<br/><br/>If it's not showing, 
      it's because the box is collapsed - you can open it by clicking the 'plus'
      button on the right.",
      "This box contains the predicted repression and number of binding sites 
      for all transcripts. You can reorder it anyway you like, and filter or 
      remove any column.<br/><br/>You can also double-click on one of the row 
      to get the details and visualize the individual binding sites - this would
      automatically move us to the transcript-centered section of the app.",
      "Moving on to the <b>transcript-centered</b> section!<br/><br/>
      The '<i>Search in gene/sequence</i>' menu on the left provides access to 
      sub-sections, the first of which ('subject') is about selecting the 
      sequence we want to scan.",
      "There are two ways to select a sequence to scan, available through the 
      tabs on the top right:<br>
      you can either scan an annotated transcript, or a custom sequence that 
      you'll copy-paste. We'll look at the former.",
      "In this sub-tab, you can first select the gene and then select the 
      transcript.<br/>Again, you can delete the content of the box and type the 
      first few letters to get the matching options.<br/>
      An overview of the selected sequence is then shown at the bottom.",
      "The next step is to select the miRNA(s) for which you want to scan 
      binding sites.<br/><br/>
      You can click in the box again and type the beginning of the miRNA(s) you
      wish to add and the matching miRNAs will show up. Alternatively, you can 
      select a set of miRNAs using the buttons.",
      "Finally, you can click on the scan button to launch it!<br/>
      If it's disabled, it's most likely because you didn't select any miRNA 
      and/or sequence.",
      "The putative binding sites are shown in the <i>'hits'</i> tab.<br/>
      You've got two ways to browse sites: you can either visualize them on a 
      plot along the length of the sequence, or through the table.<br/>
      In both cases, you can <i>double click</i> on a given match to view the 
      supplementary 3' alignment on the target sequence.",
      "Well, that's about it for the basic functions! Please let us know if you
      encounter problems!"
    )
  )
}

.getHelpModal <- function(topic){
  switch(topic,
    hitsCol=showModal(modalDialog(title="Columns of the hits table:",
      tags$p("The start and end columns represent the coordinates of the seed
binding site (i.e. corresponding to positions 1-8 of the miRNA) on the
target sequence. The coordinates are based on the beginning of the sequence
scanned (i.e. if you scanned only the UTR region, 1 represents the first
nucleotide of the UTR)."), tags$p("
The 'type' column indicates the type of match in the seed region."), tags$p("
The 'log_kd' column indicates the log of the estimated dissociation rate. A
lower log_kd value is indicative of a stronger affinity."), tags$p("
The 'p3.score' represent the strength of the 3' supplementary alignment. In
most cases, it roughly corresponds to the number of consecutive complementary
bases (exluding the seed region). However, because of tolerance for G-U
bindings or mismatches in large regions of complementarity, it can depart from
this rule of thumb. To visualize this alignment, you can simply double click
on one of the rows of the table."), tags$p("
The 'note' column contains eventual special features of the binding site (e.g.
prediction of TDMD sites)."), easyClose=TRUE, footer=NULL)),
    stypes=showModal(modalDialog(easyClose=TRUE, footer=NULL, size="l",
      title="Description of the site types",
      tags$p("miRNAs can bind to mRNAs in different kinds, generally defined by
        the 'binding site types'. See the figure below from ", 
      tags$a(href="https://doi.org/10.1016/j.cell.2009.01.002", 
             "Bartel (2009)", target="_blank"), "for a first overview and 
        the corresponding  review for detailed information."),
      tags$div(style="height: 576px;", imageOutput("bartel2009")), 
      tags$p("In brief, there are canonical and non-canonical binding sites. The
        canonical binding sites are primarily classified by the number of 
        consecutive nucleotides of the extended seed region of a miRNA that are 
        bound to their complements. '7mer-m8' indicates that the 8th nucleotide 
        (nt) of the miRNA is bound at that binding site, '7mer-A1' that there 
        is an 'A' opposite the first nt of the miRNA, which is beneficial at 
        that specific position. Non-canoncial binding sites typically include 
        parts of the seed, sometimes with wobble bindings or bulged out 
        nucleotides, but often do not cause effective silencing of transcripts.
        An overview of the affinity of the top sites can be seen in the miRNA
        tab."),
      tags$p(tags$a(href="https://dx.doi.org/10.1126/science.aav1741", 
        "McGeary, Lin et al. (2019)", target="_blank"), " show that in order to
        determine the efficacy of a binding site, it's important to not only 
        consider the type of the site but also take the flanking di-nucleotides
        into account. Finally, in some cases supplementary binding on the miRNA
        3' side can increase the stability of the binding or lead to other
        behaviors (such as target-mediated miRNA degradataion), however the 
        impact of such supplementary binding is difficult to predict.")
    )),
    manhattan=showModal(modalDialog(easyClose=TRUE, footer=NULL,
      title="Plotting sites along transcript",
      tags$p("This plot shows the predicted miRNA binding sites at their 
        positions (on the x-axis) within the selected sequence. The '-log_kd' 
        (= -log(KD)) information on the y-axis is a proxy for the  affinity of 
        a specific miRNA to its complementary sequence at that position. The 
        higher a '-log_kd' value is, the more time the Ago-miRNA complex will 
        spend at that site and the more likely this particular binding site 
        will lead to post-transcriptional repression of the transcript."),
      tags$p("If the scan has been performed with a single miRNA, a green area 
        highlighting will show the affinity range of the strongest site types 
        for this miRNA (i.e. the -log_kd range of 8mers), which can help assess
        the strength of individual binding sites."),
      tags$p("If the scan has been performed with several miRNAs, by default 
        only those with the strongest sites (see also the order in the 
        hits-table) will be displayed in the plot. The number of shown miRNAs 
        can be altered in the 'Number of top miRNAs displayed' input"),
      tags$p("A 'log_kd' of -1 (indicated by the dashed red line) indicates 
             background binding."),
      tags$p("Note log_kd values are not necessarily comparable across miRNAs!")
    )),
    collections=showModal(modalDialog(easyClose=TRUE, footer=NULL,
      title="miRNA Kd-Model collections",
      tags$p("In this section you can select a miRNA species collection, which 
        will determine the available miRNA-KdModels that you can subsequently 
        choose to scan sequences or display in the miRNA-based tab. We 
        pre-compiled KdModels for all human, mouse and rat",
        tags$a(href="https://www.mirbase.org/index.shtml", "miRBase v22", 
          target="_blank"), "miRNAs. To get an overview of the number of 
        available miRNAs for the selected species and their",
        tags$a(href="http://www.targetscan.org/vert_80/", "Targetscan", 
          target="_blank"),"conservation status simply click on 'Details'."),
      tags$p("You can also run the App with your own list of KdModels if your 
        species of interest is for example not provided by default. Consult the
        ", tags$a(href="http://bioconductor.org/packages/release/bioc/html/scanMiRApp.html", 
        "vignette", target="_blank"),"of the Bioconductor package for further 
        information.")
    )),
    modalDialog(title=topic, "No help currently available for this topic.")
  )
}

---
title: "scanMiRApp: shiny app and related convenience functions"
author: 
- name: Pierre-Luc Germain
  affiliation:
    - D-HEST Institute for Neuroscience, ETH
    - Lab of Statistical Bioinformatics, UZH
- name: Michael Soutschek
  affiliation: Lab of Systems Neuroscience, D-HEST Institute for Neuroscience, ETH
- name: Fridolin Groß
  affiliation: Lab of Systems Neuroscience, D-HEST Institute for Neuroscience, ETH
package: scanMiRApp
output:
  BiocStyle::html_document
abstract: |
  Covers the creation of ScanMiRAnno objects, setting up the shiny app, and 
  using the wrappers.
vignette: |
  %\VignetteIndexEntry{scanMiRApp}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

`ScanMiRApp` offers a `r CRANpkg("shiny")` interface to the `scanMiR` package,
as well as convenience function to simplify its use with common annotations.

# ScanMiRAnno objects

Both the shiny app and the convenience functions rely on objects of the class 
`ScanMiRAnno`, which contain the different pieces of annotation relating to a 
species and genome build. Annotations for human (GRCh38), mouse (GRCm38) and 
rat (Rnor_6) can be obtained as follows:

```{r}
library(scanMiRApp)
# anno <- ScanMiRAnno("Rnor_6")
# for this vignette, we'll work with a lightweight fake annotation:
anno <- ScanMiRAnno("fake")
anno
```

You can also build your own `ScanMiRAnno` object by providing the function with
the different components (minimally, a `r Biocpkg("BSgenome")` and an 
`r Biocpkg("ensembldb")` object - see `?ScanMiRAnno` for more information). For
minimal functioning with the shiny app, the `models` slot additionally needs to
be populated with a `KdModelList` (see the corresponding vignette of the 
`scanMiR` package for more information).

In addition, `ScanMiRAnno` objects can contain pre-compiled scans and 
aggregations, which are especially meant to speed up the shiny application. 
These should be saved as [IndexedFst](IndexedFST.html) files and should be 
respectively indexed by transcript and by miRNA, and stored in the `scan` and 
`aggregated` slot of the object.

<br/><br/>

# Convenience functions

## Obtaining the UTR sequence of a transcript

The transcript (or UTR) sequence for any (set of) transcript(s) in the 
annotation can be obtained with:

```{r}
seq <- getTranscriptSequence("ENSTFAKE0000056456", anno)
seq
```

## Plotting sites on the UTR sequence of a transcript

Binding sites of a given miRNA on a transcript can be visualized with:

```{r}
plotSitesOnUTR(tx="ENSTFAKE0000056456", annotation=anno, miRNA="hsa-miR-155-5p")
```

This will fetch the sequence, perform the scan, and plot the results.

## Running a full-transcriptome scan

The `runFullScan` function can be used to launch a the scan for all miRNAs on 
all protein-coding transcripts (or their UTRs) of a genome. These scans can then
be used to speed up the shiny app (see below). They can simply be launched as:

```{r}
m <- runFullScan(anno)
m
```

Multi-threading can be enabled through the `ncores` argument. See `?runFullScan`
for more options.

## Detecting enriched miRNA-target pairs

The `enrichedMirTxPairs` identifies miRNA-target enrichments (which could 
indicate sponge- or cargo-like behaviors) by means of a binomial model 
estimating the probability of the given number of binding sites for a given 
pair given the total number of bindings sites for the miRNA (across all 
transcripts) and transcript (across all miRNAs) in question. The output is
a data.frame indicating, for each pair passing some lenient filtering, the 
transcript, miRNA, the number of 7mer/8mer sites, and the binomial log(p-value)
of the combination. We strongly recommend further filtering this list by 
expression of the transcript in the system of interest, otherwise some 
transcripts with very low expression (and hence biologically irrelevant) might 
come up as strongly enriched.

<br/><br/>

# Shiny app

The features of the shiny app are organized into two main components:

* transcript (or sequence) -centered features are available in the 
_search in gene/sequence_ tab. These for instance allow to scan custom 
sequences or selected transcript sequences for miRNA binding sites, visualize 
them on the transcript, and visualize the sequence pairing of specific matches.

* the miRNA-centered features are available in the _miRNA-based_ tab. It shows
the general binding specificity of a given miRNA. If the `scanMiRAnno` object
contained aggregated data (see below), the tab also shows the top predicted 
targets for the miRNAs.


## Setting up the application

A `ScanMiRAnno` object is the minimal input for the shiny app, and multiple such
objects can be provided in the form of a named list:

```{r, eval=FALSE}
scanMiRApp( list( nameOfAnnotation=anno ) )
```

Launched with this object, the app will not have access to any pre-compiled 
scans or to aggregated data. This means that scans will be performed on the fly,
which also means that they will be slower. In addition, it means that the top
targets based on aggregated repression estimates (in the _miRNA-based_ tab) 
will not be available. To provide this additional information, you first need to
prepare the objects as [IndexedFst](IndexedFST.html) files. Assuming you've 
saved (or downloaded) the scans as `scan.rds` and the aggregated data as 
`aggregated.rds`, you can re-save them as `IndexedFst` (here in the folder 
`out_path`) and add them to the `anno` object as follows:

```{r, eval=FALSE}
# not run
anno <- ScanMiRAnno("Rnor_6")
saveIndexedFst(readRDS("scan.rds"), "seqnames", file.prefix="out_path/scan")
saveIndexedFst(readRDS("aggregated.rds"), "miRNA", 
               file.prefix="out_path/aggregated")
anno$scan <- loadIndexedFst("out_path/scan")
anno$aggregated <- loadIndexedFst("out_path/aggregated")
# then launch the app
scanMiRApp(list(Rnor_6=anno))
```

The same could be done for multiple ScanMiRAnno objects. If `scanMiRApp` is 
launched without any `annotation` argument, it will generate anno objects for
the three base species (without any pre-compiled data).

## Multi-threading

Multithreading can be enabled in the shiny app by calling `scanMiRApp()` (or 
the underlying `scanMiRserver()`) with the `BP` argument, e.g.:
```
scanMiRApp(..., BP=BiocParallel::MulticoreParam(ncores))
```

where `ncores` is the number of threads to use. This will enable 
multi-threading for the scanning functions, which makes a big difference when 
scanning for many miRNAs at a time. In addition, multi-threading can be
used to read the `IndexedFst` files, which is enabled by the `nthreads` of the
`loadIndexedFst` function. However, since reading is quite fast already with a 
single core, improvements there are typically fairly marginal.

## Caching

By default, the app has a caching system which means that if a user wants to 
launch the same scan with the same parameters twice, the results will be 
re-used instead of re-computed. The cache has a maximum size (by default 10MB)
per user, beyond which older cache items will be removed. The cache size can be
manipulated through the `maxCacheSize` argument.


# Session info {.unnumbered}

```{r sessionInfo, echo=FALSE}
sessionInfo()
```

---
title: "The IndexedFst class"
author: Pierre-Luc Germain
package: scanMiRApp
output:
  BiocStyle::html_document
abstract: |
  Indexed wrapper for fst objects, enabling indexed fast random access to on-disk data.frames or GRanges
vignette: |
  %\VignetteIndexEntry{IndexedFst}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include=FALSE}
library(BiocStyle)
```

# IndexedFst

The `IndexedFst` class provides fast named random access to indexed `fst` files. It is based on the `r CRANpkg("fst")` package, which provides fast random reading of data frames. This is particularly useful to manipulate large collections of binding sites without loading them all in memory.

Creating an indexed fst file from a data.frame is very simple:

```{r}
library(scanMiRApp)
# we create a temporary directory in which the files will be saved
tmp <- tempdir()
f <- file.path(tmp, "test")
# we create a dummy data.frame
d <- data.frame( category=sample(LETTERS[1:4], 10000, replace=TRUE),
                 var2=sample(LETTERS, 10000, replace=TRUE),
                 var3=runif(10000) )

saveIndexedFst(d, index.by="category", file.prefix=f)
```

The file can then be loaded (without having all the data in memory) in the following way:

```{r}
d2 <- loadIndexedFst(f)
class(d2)
summary(d2)
```

We can see that `d2` is considerably smaller than the original `d`:

```{r}
format(object.size(d),units="Kb")
format(object.size(d2),units="Kb")
```

Nevertheless, a number of functions can be used normally on the object:

```{r}
nrow(d2)
ncol(d2)
colnames(d2)
head(d2)
```

In addition, the object can be accessed as a list (using the indexed variable). Since in this case the file is indexed using the category column, the different categories can be accessed as `names` of the object:

```{r}
names(d2)
lengths(d2)
```
We can read specifically the rows pertaining to one category using:

```{r}
catB <- d2$B
head(catB)
```

# Storing GRanges as IndexedFst

In addition to data.frames, [GRanges](https://bioconductor.org/packages/release/bioc/html/GenomicRanges.html) can be saved as indexed Fst. To demonstrate this, we first create a dummy GRanges object:

```{r}
library(GenomicRanges)
gr <- GRanges(sample(LETTERS[1:3],200,replace=TRUE), IRanges(seq_len(200), width=2))
gr$propertyA <- factor(sample(letters[1:5],200,replace=TRUE))
gr
```
Again the file can then be loaded (without having all the data in memory) in the following way:

```{r}
f2 <- file.path(tmp, "test2")
saveIndexedFst(gr, index.by="seqnames", file.prefix=f2)
d1 <- loadIndexedFst(f2)
names(d1)
head(d1$A)
```
Similarly, we could index using a different column:

```{r}
saveIndexedFst(gr, index.by="propertyA", file.prefix=f2)
d2 <- loadIndexedFst(f2)
names(d2)
```

# More...

## Multithreading

The `r CRANpkg("fst")` package supports multithreaded reading and writing. This can also be applied for `IndexedFst`, using the `nthreads` argument of `loadIndexedFst` and `saveIndexedFst`.

## Under the hood

The `IndexedFst` class is simply a wrapper around the `fst` package. In addition to the `fst` file, an `rds` file is saved containing the index data. For example, for our last example, the following files have been saved:

```{r}
list.files(tmp, "test2")
```
Either file (or the prefix) can be used for loading, but both files need to have the same prefix.


<br/><br/>

# Session info {.unnumbered}

```{r sessionInfo, echo=FALSE}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getTranscriptSequence.R
\name{plotSitesOnUTR}
\alias{plotSitesOnUTR}
\title{plotSitesOnUTR}
\usage{
plotSitesOnUTR(
  tx,
  annotation,
  miRNA = NULL,
  label_6mers = FALSE,
  label_notes = FALSE,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{tx}{An ensembl TranscriptID}

\item{annotation}{A \code{\link{ScanMiRAnno}} object.}

\item{miRNA}{A miRNA name in the mirbase format (eg. "hsa-miR-485-5p"), a
`KdModel`, or a miRNA sequence or target seed.}

\item{label_6mers}{Logical whether to label 6mer sites in the plot}

\item{label_notes}{Logical whether to label special sites in the plot (as
TDMD or Slicing)}

\item{verbose}{Logical; whether to print updates on the processing}

\item{...}{Any further arguments passed to
\code{\link[scanMiR]{findSeedMatches}}}
}
\value{
Returns a ggplot.
}
\description{
Wrapper function with minimal arguments to plot scanMiR-Binding
sites on 3'UTRs of specified transcripts. The red dashed line indicates the
background threshhold is indicated, the lightblue dashed line shows the
average 8mer dissociation rate of the given miRNA
}
\examples{
anno <- ScanMiRAnno("fake")
plotSitesOnUTR( tx="ENSTFAKE0000056456", annotation=anno,
                miRNA="hsa-miR-155-5p" )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IndexedFst.R
\docType{class}
\name{IndexedFst-class}
\alias{IndexedFst-class}
\alias{IndexedFst}
\alias{show,IndexedFst-method}
\alias{summary,IndexedFst-method}
\alias{names,IndexedFst-method}
\alias{length,IndexedFst-method}
\alias{lengths,IndexedFst-method}
\alias{nrow,IndexedFst-method}
\alias{ncol,IndexedFst-method}
\alias{colnames,IndexedFst-method}
\alias{[[,IndexedFst,ANY,ANY-method}
\alias{[,IndexedFst,ANY,ANY,ANY-method}
\alias{$,IndexedFst-method}
\alias{head,IndexedFst-method}
\alias{as.data.frame,IndexedFst-method}
\title{IndexedFst}
\usage{
\S4method{show}{IndexedFst}(object)

\S4method{summary}{IndexedFst}(object)

\S4method{names}{IndexedFst}(x)

\S4method{length}{IndexedFst}(x)

\S4method{lengths}{IndexedFst}(x)

\S4method{nrow}{IndexedFst}(x)

\S4method{ncol}{IndexedFst}(x)

\S4method{colnames}{IndexedFst}(x)

\S4method{[[}{IndexedFst,ANY,ANY}(x, i, j = NULL, ...)

\S4method{[}{IndexedFst,ANY,ANY,ANY}(x, i, j = NULL, ..., drop = TRUE)

\S4method{$}{IndexedFst}(x, name)

\S4method{head}{IndexedFst}(x, n = 6L, ...)

\S4method{as.data.frame}{IndexedFst}(x, name)
}
\arguments{
\item{object}{an IndexedFst object}

\item{x}{an IndexedFst object}

\item{i}{the desired index (either numeric or name)}

\item{j, drop}{ignored}

\item{...}{ignored}

\item{name}{the indexed name to fetch}

\item{n}{the desired number of rows}
}
\value{
Depends on the method
}
\description{
Objects of the IndexedFst class enable fast named random access to FST files.
This is particularly appropriate for large data.frames which often need to
 be accessed according to the (e.g. factor) value of a particular column.
}
\examples{
# we first create and save an indexed FST file
tmp <- tempdir()
f <- system.file(tmp, "test")
d <- data.frame( category=sample(LETTERS[1:4], 10000, replace=TRUE),
                 var2=sample(LETTERS, 10000, replace=TRUE),
                 var3=runif(10000) )
format(object.size(d),units="Kb")
saveIndexedFst(d, "category", f)
rm(d)
# we then load the index, and can use category names for random access:
d <- loadIndexedFst(f)
format(object.size(d),units="Kb")
nrow(d)
names(d)
head(d$A)
}
\seealso{
\code{\link{saveIndexedFst}}, \code{\link{loadIndexedFst}}
}
\author{
Pierre-Luc Germain, \email{pierre-luc.germain@hest.ethz.ch}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{fakeTxDb}
\alias{fakeTxDb}
\title{Example 'fake' TxDb object}
\value{
a named character vector of length 1.
}
\description{
A fake transcript database used for examples.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrichedPairs.R
\name{enrichedMirTxPairs}
\alias{enrichedMirTxPairs}
\title{enrichedMirTxPairs}
\usage{
enrichedMirTxPairs(m, minSites = 5, max.binom.p = 0.001)
}
\arguments{
\item{m}{A GRanges of matches, as produced by \code{\link{findSeedMatches}}.
This will be filtered down to only 8mer and 7mer sites.}

\item{minSites}{The minimum number of sites for a given miRNA-target pair to
be considered.}

\item{max.binom.p}{The maximum binomial p-value of miRNA-target pairs.}
}
\value{
A data.frame of top combinations, including number of sites and
the log-transformed binomial p-value.
}
\description{
Identifies pairs of miRNA and target transcripts that have an unexpectedly
high number of sites.
}
\examples{
# we create a dummy scan (see `runFullScan`)
library(scanMiR)
seqs <- getRandomSeq(n=10)
mirs <- c("TTGTATAA","AGCATTAA")
m <- findSeedMatches(seqs,mirs,verbose=FALSE)
# we look for enriched pairs
res <- enrichedMirTxPairs(m, minSites=1, max.binom.p=1)
res
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getTranscriptSequence.R
\name{getTranscriptSequence}
\alias{getTranscriptSequence}
\title{getTranscriptSequence}
\usage{
getTranscriptSequence(
  tx = NULL,
  annotation,
  annoFilter = NULL,
  extract = c("UTRonly", "withORF", "exons"),
  ...
)
}
\arguments{
\item{tx}{The ensembl ID of the transcript(s)}

\item{annotation}{A \code{\link{ScanMiRAnno}} object.}

\item{annoFilter}{An optional `AnnotationFilter` or `AnnotationFilterList` to
further filter the set of transcripts to be extracted}

\item{extract}{Which parts of the transcripts to extract. For `UTRonly`
(default) only the 3' UTR regions are extracted, `withORF` additionally
extracts the coding regions, and `exons` extracts all exons}

\item{...}{Passed to \code{\link{AnnotationHub}}}
}
\value{
A \code{\link{DNAStringSet}}.
}
\description{
Utility wrapper to extracts the sequence of a given transcript (UTR or
CDS+UTR).
}
\examples{
anno <- ScanMiRAnno("fake")
seq <- getTranscriptSequence( tx="ENSTFAKE0000056456", annotation=anno )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scanMiRAnno.R
\name{ScanMiRAnno-methods}
\alias{ScanMiRAnno-methods}
\alias{summary,ScanMiRAnno-method}
\alias{show,ScanMiRAnno-method}
\title{Methods for the \code{\link{ScanMiRAnno}} class}
\usage{
\S4method{summary}{ScanMiRAnno}(object)

\S4method{show}{ScanMiRAnno}(object)
}
\arguments{
\item{object}{An object of class \code{\link{ScanMiRAnno}}}
}
\value{
Depends on the method.
}
\description{
Methods for the \code{\link{ScanMiRAnno}} class
}
\seealso{
\code{\link{ScanMiRAnno}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scanMiRui.R
\name{scanMiRui}
\alias{scanMiRui}
\title{scanMiRui}
\usage{
scanMiRui()
}
\value{
A shiny ui
}
\description{
UI for the scanMiR app.
}
\examples{
ui <- scanMiRui()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scanMiRAnno.R
\docType{class}
\name{ScanMiRAnno-class}
\alias{ScanMiRAnno-class}
\alias{ScanMiRAnno}
\title{ScanMiRAnno}
\usage{
ScanMiRAnno(
  species = NULL,
  genome = NULL,
  ensdb = NULL,
  models = NULL,
  scan = NULL,
  aggregated = NULL,
  version = NULL,
  addDBs = list(),
  ...
)
}
\arguments{
\item{species}{The species/build acronym for automatic construction; if
omitted, `genome` and `ensdb` should be given. Current possible values are:
GRCh38, GRCm38, GRCm39, Rnor_6.}

\item{genome}{A \link[BSgenome]{BSgenome-class}, or a
\code{\link[rtracklayer]{TwoBitFile}}}

\item{ensdb}{An \link[ensembldb]{EnsDb-class} (or a
\link[GenomicFeatures]{TxDb-class}) object}

\item{models}{An optional KdModelList}

\item{scan}{An optional full scan (IndexedFst or GRanges)}

\item{aggregated}{An optional per-transcript aggregation (IndexedFst or
data.frame)}

\item{version}{optional ensembl version}

\item{addDBs}{A named list of additional tx-miRNA databases, each of which
should be a data.frame with the columns 'transcript', 'miRNA', and 'score'.}

\item{...}{Arguments passed to `AnnotationHub`}
}
\value{
A `ScanMiRAnno` object
}
\description{
ScanMiRAnno
}
\examples{
anno <- ScanMiRAnno(species="fake")
anno
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IndexedFst.R
\name{loadIndexedFst}
\alias{loadIndexedFst}
\alias{saveIndexedFst}
\title{Saving and loading IndexedFst}
\usage{
loadIndexedFst(file, nthreads = 1)

saveIndexedFst(
  d,
  index.by,
  file.prefix,
  nthreads = 1,
  index.properties = NULL,
  add.info = list(),
  ...
)
}
\arguments{
\item{file}{Path to the fst file, it's index (.idx), or their prefix.}

\item{nthreads}{Number of threads to use for reading (default 1). This does
not affect the loading of the index itself, but will affect all downstream
reading operations performed on the object. If NULL, will use
`fst::threads_fst()`.}

\item{d}{A data.frame or \code{\link[GenomicRanges]{GRanges}} object}

\item{index.by}{A column of `d` by which it should be indexed.}

\item{file.prefix}{Path and prefix of the output files.}

\item{index.properties}{An optional data.frame of properties, with the levels
of `index.by` as row names.}

\item{add.info}{An optional list of additional information to save.}

\item{...}{Passed to `write.fst`}
}
\value{
`loadIndexedFst` returns an object of class
\code{\link{IndexedFst-class}}, and `saveIndexedFst` returns nothing.
}
\description{
Functions to save or load and indexed \code{\link[fst]{fst}} file

Saves a data.frame (or GRanges object) into an indexed FST file.
}
\examples{
# we first create and save an indexed FST file
tmp <- tempdir()
f <- system.file(tmp, "test")
d <- data.frame( category=sample(LETTERS[1:4], 10000, replace=TRUE),
                 var2=sample(LETTERS, 10000, replace=TRUE),
                 var3=runif(10000) )
saveIndexedFst(d, "category", f)
# we then load the index, and can use category names for random access:
d <- loadIndexedFst(f)
}
\seealso{
\code{\link{IndexedFst-class}}

\code{\link{IndexedFst-class}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runFullScan.R
\name{runFullScan}
\alias{runFullScan}
\title{runFullScan}
\usage{
runFullScan(
  annotation,
  mods = NULL,
  annoFilter = NULL,
  extract = c("UTRonly", "withORF", "exons"),
  onlyCanonical = TRUE,
  shadow = 15,
  cores = 1,
  maxLogKd = c(-1, -1.5),
  save.path = NULL,
  ...
)
}
\arguments{
\item{annotation}{A \code{\link{ScanMiRAnno}} object}

\item{mods}{An optional `KdModelList` (defaults to the one in `annotation`)}

\item{annoFilter}{An optional `AnnotationFilter` or `AnnotationFilterList` to
filter the set of transcripts to be extracted}

\item{extract}{Which parts of the transcripts to extract. For `UTRonly`
(default) only the 3' UTR regions are extracted, `withORF` additionally
extracts the coding regions, and `exons` extracts all exons}

\item{onlyCanonical}{passed to \code{\link[scanMiR]{findSeedMatches}}}

\item{shadow}{The size of the ribosomal shadow at the UTR starts}

\item{cores}{The number of threads to use. Alternatively accepts a
\code{\link[BiocParallel]{BiocParallelParam-class}}, as for instance produced
by \code{\link[BiocParallel]{MulticoreParam}}.}

\item{maxLogKd}{The maximum log_kd of sites to report}

\item{save.path}{Optional, the path to which to save the results}

\item{...}{Arguments passed to \code{\link[scanMiR]{findSeedMatches}}}
}
\value{
A `GRanges` object
}
\description{
Runs a full miRNA scan on all protein-coding transcripts (or UTRs) of an
annotation.
}
\examples{
anno <- ScanMiRAnno("fake")
m <- runFullScan( annotation=anno )
m
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scanMiRserver.R
\name{scanMiRApp}
\alias{scanMiRApp}
\title{scanMiRApp
A wrapper for launching the scanMiRApp shiny app}
\usage{
scanMiRApp(annotations = NULL, ...)
}
\arguments{
\item{annotations}{A named list of \code{\link{ScanMiRAnno}} objects. If
omitted, will use the base ones.}

\item{...}{Passed to \code{\link{scanMiRserver}}}
}
\value{
A shiny app
}
\description{
scanMiRApp
A wrapper for launching the scanMiRApp shiny app
}
\examples{
if(interactive()){
  anno <- ScanMiRAnno("fake")
  scanMiRApp(list(fakeAnno=anno))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scanMiRserver.R
\name{scanMiRserver}
\alias{scanMiRserver}
\title{scanMiRserver}
\usage{
scanMiRserver(
  annotations = list(),
  modlists = NULL,
  gc.time = Inf,
  maxCacheSize = 10 * 10^6,
  BP = SerialParam()
)
}
\arguments{
\item{annotations}{A named list of \code{\link{ScanMiRAnno}} object.}

\item{modlists}{A named list of `KdModelList` objects. If omitted, will
fetch it from the annotation objects.}

\item{gc.time}{The number of microseconds after which a garbage collect
should be performed.}

\item{maxCacheSize}{Maximum cache size in bytes.}

\item{BP}{BPPARAM for multithreading}
}
\value{
A shiny server function
}
\description{
Server function for the scanMiR shiny app. Most users are expected to use
\code{\link{scanMiRApp}} instead.
}
\examples{
# we'd normally fetch a real annotation:
# anno <- ScanMiRAnno("Rnor_6")
# here we'll use a fake one:
anno <- ScanMiRAnno("fake")
srv <- scanMiRserver(list(fake=anno))
}
